"""Nf-tower related constants."""
from enum import StrEnum


class NfTowerStatus(StrEnum):
    """NF-Tower job submission status."""

    SUBMITTED: str = "SUBMITTED"
    RUNNING: str = "RUNNING"
    SUCCEEDED: str = "SUCCEEDED"
    FAILED: str = "FAILED"
    CANCELLED: str = "CANCELLED"
    UNKNOWN: str = "UNKNOWN"



"""Post-processing Demultiplex API."""
import logging
from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.exc import FlowCellError, MissingFilesError
from cg.meta.demultiplex.housekeeper_storage_functions import (
    store_flow_cell_data_in_housekeeper,
)
from cg.meta.demultiplex.status_db_storage_functions import (
    store_flow_cell_data_in_status_db,
    store_sample_data_in_status_db,
    store_sequencing_metrics_in_status_db,
)
from cg.meta.demultiplex.utils import create_delivery_file_in_flow_cell_directory
from cg.meta.demultiplex.validation import is_flow_cell_ready_for_postprocessing
from cg.models.cg_config import CGConfig
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.store import Store

LOG = logging.getLogger(__name__)


class DemuxPostProcessingAPI:
    """Post demultiplexing API class."""

    def __init__(self, config: CGConfig) -> None:
        self.flow_cells_dir: Path = Path(config.flow_cells_dir)
        self.demultiplexed_runs_dir: Path = Path(config.demultiplexed_flow_cells_dir)
        self.status_db: Store = config.status_db
        self.hk_api: HousekeeperAPI = config.housekeeper_api
        self.dry_run: bool = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Set dry run."""
        LOG.debug(f"Set dry run to {dry_run}")
        self.dry_run = dry_run

    def finish_flow_cell(
        self,
        flow_cell_directory_name: str,
        bcl_converter: str | None = None,
        force: bool = False,
    ) -> None:
        """Store data for the demultiplexed flow cell and mark it as ready for delivery.
        This function:
            - Stores the flow cell in the status database
            - Stores sequencing metrics in the status database
            - Updates sample read counts in the status database
            - Stores the flow cell data in the housekeeper database
            - Creates a delivery file in the flow cell directory
        Args:
            flow_cell_directory_name (str): The name of the flow cell directory to be finalized.
            bcl_converter (str): The name of the bcl converter used for demultiplexing.
            force (bool): If True, the flow cell will be finalized even when it is already marked for delivery.
        Raises:
            FlowCellError: If the flow cell directory or the data it contains is not valid.
        """
        if self.dry_run:
            LOG.info(f"Dry run will not finish flow cell {flow_cell_directory_name}")
            return

        LOG.info(f"Finish flow cell {flow_cell_directory_name}")

        flow_cell_out_directory = Path(self.demultiplexed_runs_dir, flow_cell_directory_name)

        flow_cell = FlowCellDirectoryData(
            flow_cell_path=flow_cell_out_directory, bcl_converter=bcl_converter
        )

        sample_sheet_path: Path = self.hk_api.get_sample_sheet_path(flow_cell.id)
        flow_cell.set_sample_sheet_path_hk(hk_path=sample_sheet_path)
        LOG.debug("Set path for Housekeeper sample sheet in flow cell")

        try:
            is_flow_cell_ready_for_postprocessing(
                flow_cell_output_directory=flow_cell_out_directory,
                flow_cell=flow_cell,
                force=force,
            )
        except (FlowCellError, MissingFilesError) as e:
            LOG.warning(f"Flow cell {flow_cell_directory_name} will be skipped: {e}")
            return

        try:
            self.store_flow_cell_data(flow_cell)
        except Exception as e:
            LOG.error(f"Failed to store flow cell data: {str(e)}")
            raise
        create_delivery_file_in_flow_cell_directory(flow_cell_out_directory)

    def finish_all_flow_cells(self) -> bool:
        """Finish all flow cells that need it."""
        flow_cell_dirs = self.get_all_demultiplexed_flow_cell_dirs()
        is_error_raised: bool = False
        for flow_cell_dir in flow_cell_dirs:
            try:
                self.finish_flow_cell(flow_cell_dir.name)
            except Exception as error:
                LOG.error(f"Failed to finish flow cell {flow_cell_dir.name}: {str(error)}")
                is_error_raised = True
                continue
        return is_error_raised

    def store_flow_cell_data(self, parsed_flow_cell: FlowCellDirectoryData) -> None:
        """Store data from the flow cell directory in status db and housekeeper."""
        store_flow_cell_data_in_status_db(
            parsed_flow_cell=parsed_flow_cell,
            store=self.status_db,
        )
        store_sequencing_metrics_in_status_db(flow_cell=parsed_flow_cell, store=self.status_db)
        store_sample_data_in_status_db(flow_cell=parsed_flow_cell, store=self.status_db)
        store_flow_cell_data_in_housekeeper(
            flow_cell=parsed_flow_cell,
            hk_api=self.hk_api,
            flow_cell_run_dir=self.flow_cells_dir,
            store=self.status_db,
        )

    def get_all_demultiplexed_flow_cell_dirs(self) -> list[Path]:
        """Return all demultiplex flow cell out directories."""
        demultiplex_flow_cells: list[Path] = []
        for flow_cell_dir in self.demultiplexed_runs_dir.iterdir():
            if flow_cell_dir.is_dir():
                LOG.debug(f"Found directory {flow_cell_dir}")
                demultiplex_flow_cells.append(flow_cell_dir)
        return demultiplex_flow_cells



"""Fixtures for CLI tests."""
from pathlib import Path
from typing import Any

import pytest
from click.testing import CliRunner

from cg.apps.crunchy import CrunchyAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.meta.compress import CompressAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from tests.cli.compress.conftest import CaseInfo
from tests.store_helpers import StoreHelpers


@pytest.fixture
def cli_runner() -> CliRunner:
    """Create a CliRunner"""
    return CliRunner()


@pytest.fixture
def application_tag() -> str:
    """Return a dummy tag"""
    return "dummy_tag"


@pytest.fixture
def base_context(
    base_store: Store, housekeeper_api: HousekeeperAPI, cg_config_object: CGConfig
) -> CGConfig:
    """context to use in CLI."""
    cg_config_object.status_db_ = base_store
    cg_config_object.housekeeper_api_ = housekeeper_api
    return cg_config_object


@pytest.fixture
def before_date() -> str:
    """Return a before date string"""
    return "1999-12-31"


@pytest.fixture
def disk_store(base_context: CGConfig) -> Store:
    """context to use in cli"""
    return base_context.status_db


class MockCompressAPI(CompressAPI):
    """Mock out necessary functions for running the compress CLI functions."""

    def __init__(self):
        """initialize mock"""
        super().__init__(hk_api=None, crunchy_api=None, demux_root="")
        self.ntasks = 12
        self.mem = 50
        self.fastq_compression_success = True
        self.spring_decompression_success = True
        self.dry_run = False

    def set_dry_run(self, dry_run: bool):
        """Update dry run."""
        self.dry_run = dry_run

    def compress_fastq(self, sample_id: str, dry_run: bool = False):
        """Return if compression was successful."""
        _ = sample_id, dry_run
        return self.fastq_compression_success

    def decompress_spring(self, sample_id: str, dry_run: bool = False):
        """Return if decompression was succesful."""
        _ = sample_id, dry_run
        return self.spring_decompression_success


@pytest.fixture
def compress_api():
    """Return a compress API context."""
    return MockCompressAPI()


@pytest.fixture
def compress_context(
    compress_api: CompressAPI, store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a compress context."""
    cg_config_object.meta_apis["compress_api"] = compress_api
    cg_config_object.status_db_ = store
    return cg_config_object


@pytest.fixture
def compress_case_info(
    case_id,
    family_name,
    timestamp,
    later_timestamp,
    wgs_application_tag,
):
    """Returns a object with information about a case."""
    return CaseInfo(
        case_id=case_id,
        family_name=family_name,
        timestamp=timestamp,
        later_timestamp=later_timestamp,
        application_tag=wgs_application_tag,
    )


@pytest.fixture
def populated_compress_store(store, helpers, compress_case_info, analysis_family):
    """Return a store populated with a completed analysis."""
    # Make sure that there is a case where analysis is completed
    helpers.ensure_case_from_dict(
        store,
        case_info=analysis_family,
        app_tag=compress_case_info.application_tag,
        ordered_at=compress_case_info.timestamp,
        completed_at=compress_case_info.later_timestamp,
    )
    return store


@pytest.fixture
def populated_compress_context(
    compress_api: CompressAPI, populated_compress_store: Store, cg_config_object: CGConfig
) -> CGConfig:
    """Return a compress context populated with a completed analysis."""
    # Make sure that there is a case where analysis is completed
    cg_config_object.meta_apis["compress_api"] = compress_api
    cg_config_object.status_db_ = populated_compress_store
    return cg_config_object


@pytest.fixture
def real_crunchy_api(crunchy_config: dict[str, dict[str, Any]]):
    """Return Crunchy API."""
    _api = CrunchyAPI(crunchy_config)
    _api.set_dry_run(True)
    yield _api


@pytest.fixture
def real_compress_api(
    demultiplexed_runs: Path, housekeeper_api: HousekeeperAPI, real_crunchy_api: CrunchyAPI
) -> CompressAPI:
    """Return a compress API context."""
    return CompressAPI(
        crunchy_api=real_crunchy_api,
        hk_api=housekeeper_api,
        demux_root=demultiplexed_runs.as_posix(),
    )


@pytest.fixture
def real_populated_compress_fastq_api(
    real_compress_api: CompressAPI, compress_hk_fastq_bundle: dict, helpers: StoreHelpers
) -> CompressAPI:
    """Return real populated compress API."""
    helpers.ensure_hk_bundle(real_compress_api.hk_api, compress_hk_fastq_bundle)

    return real_compress_api


@pytest.fixture
def real_populated_compress_context(
    real_populated_compress_fastq_api: CompressAPI,
    populated_compress_store: Store,
    cg_config_object: CGConfig,
) -> CGConfig:
    """Return a compress context populated with a completed analysis."""
    # Make sure that there is a case where analysis is completed
    cg_config_object.meta_apis["compress_api"] = real_populated_compress_fastq_api
    cg_config_object.status_db_ = populated_compress_store
    return cg_config_object



import os
import sys
from pathlib import Path
from typing import Type

import pytest
from _pytest.fixtures import FixtureRequest

from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.cli.demultiplex.copy_novaseqx_demultiplex_data import get_latest_analysis_path
from cg.constants.demultiplexing import BclConverter, DemultiplexingDirsAndFiles
from cg.constants.sequencing import Sequencers
from cg.exc import FlowCellError
from cg.models.demultiplex.run_parameters import RunParameters
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData


def test_flow_cell_id(bcl2fastq_flow_cell_dir: Path):
    """Test parsing of flow cell id."""
    # GIVEN the path to a finished flow cell run
    # GIVEN the flow cell id
    flowcell_id: str = bcl2fastq_flow_cell_dir.name.split("_")[-1][1:]

    # WHEN instantiating a flow cell object
    flowcell_obj = FlowCellDirectoryData(flow_cell_path=bcl2fastq_flow_cell_dir)

    # THEN assert that the flow cell id is parsed
    assert flowcell_obj.id == flowcell_id


def test_flow_cell_position(bcl2fastq_flow_cell_dir: Path):
    """Test getting flow cell position."""
    # GIVEN the path to a finished flow cell
    # GIVEN a flow cell object
    flowcell_obj = FlowCellDirectoryData(flow_cell_path=bcl2fastq_flow_cell_dir)

    # WHEN fetching the flow cell position
    position = flowcell_obj.position

    # THEN assert it is A or B
    assert position in ["A", "B"]


def test_rta_exists(bcl2fastq_flow_cell: FlowCellDirectoryData):
    """Test return of RTS file."""
    # GIVEN the path to a finished flow cell
    # GIVEN a flow cell object

    # WHEN fetching the path to the RTA file
    rta_file: Path = bcl2fastq_flow_cell.rta_complete_path

    # THEN assert that the file exists
    assert rta_file.exists()


def test_get_sample_model_bcl2fastq(bcl2fastq_flow_cell: FlowCellDirectoryData):
    """Test that the sample model of a bcl2fastq flow cell is FlowCellSampleNovaSeq6000Bcl2Fastq."""
    # GIVEN a Bcl2Fastq flow cell

    # WHEN getting the sample model
    sample_model: Type[FlowCellSampleBcl2Fastq] = bcl2fastq_flow_cell.sample_type

    # THEN it is FlowCellSampleNovaSeq6000Bcl2Fastq
    assert sample_model == FlowCellSampleBcl2Fastq


def test_get_sample_model_dragen(bcl_convert_flow_cell: FlowCellDirectoryData):
    """Test that the sample model of a dragen flow cell is FlowCellSampleBCLConvert."""
    # GIVEN a dragen flow cell

    # WHEN getting the sample model
    sample_model: Type[FlowCellSampleBCLConvert] = bcl_convert_flow_cell.sample_type

    # THEN it is FlowCellSampleNovaSeq6000Bcl2Fastq
    assert sample_model == FlowCellSampleBCLConvert


def test_get_sample_model_novaseq_x(novaseq_x_flow_cell: FlowCellDirectoryData):
    """Test that the sample model of a NovaSeqX flow cell is FlowCellSampleNovaSeqX."""
    # GIVEN a NovaSeqX flow cell

    # WHEN getting the sample model
    sample_model: Type[FlowCellSampleBCLConvert] = novaseq_x_flow_cell.sample_type

    # THEN it is FlowCellSampleNovaSeq6000Bcl2Fastq
    assert sample_model == FlowCellSampleBCLConvert


def test_get_bcl_converter_by_sequencer(
    flow_cell_directory_name_demultiplexed_with_bcl2fastq: str,
):
    """Test that the bcl converter of a bcl2fastq flow cell is bcl2fastq."""
    # GIVEN a Bcl2Fastq flow cell directory

    # WHEN instantiating a flow cell object
    flow_cell = FlowCellDirectoryData(
        flow_cell_path=Path(flow_cell_directory_name_demultiplexed_with_bcl2fastq)
    )

    # THEN it sets the converter to blc2fastq
    assert flow_cell.bcl_converter == BclConverter.BCL2FASTQ


def test_flow_cell_directory_data_with_set_bcl_converter(
    flow_cell_directory_name_demultiplexed_with_bcl2fastq: str, bcl_converter=BclConverter.DRAGEN
):
    """Test that the bcl converter is set to the specified value."""
    # GIVEN a Bcl2Fastq flow cell directory

    # GIVEN the bcl converter is set to dragen

    # WHEN instantiating a flow cell object
    flow_cell = FlowCellDirectoryData(
        flow_cell_path=Path(flow_cell_directory_name_demultiplexed_with_bcl2fastq),
        bcl_converter=bcl_converter,
    )

    # THEN the bcl converter is dragen
    assert flow_cell.bcl_converter == bcl_converter


def test_flow_cell_directory_data_with_novaseq_flow_cell_directory(
    flow_cell_directory_name_demultiplexed_with_bcl_convert: str,
):
    """Test that the bcl converter is set to dragen when prodiving a novaseq flow cell directory."""
    # GIVEN a Bcl2Fastq flow cell directory

    # WHEN instantiating a flow cell object
    flow_cell = FlowCellDirectoryData(
        flow_cell_path=Path(flow_cell_directory_name_demultiplexed_with_bcl_convert),
    )

    # THEN the bcl converter is dragen
    assert flow_cell.bcl_converter == BclConverter.DRAGEN


@pytest.mark.parametrize(
    "flow_cell_fixture, expected_run_parameters_file_name",
    [
        ("hiseq_x_flow_cell", DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE),
        ("hiseq_2500_flow_cell", DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE),
    ],
)
def test_run_parameters_path(
    flow_cell_fixture: str, expected_run_parameters_file_name: str, request: FixtureRequest
):
    """."""
    # GIVEN a flow cell with a run parameters
    flow_cell: FlowCellDirectoryData = request.getfixturevalue(flow_cell_fixture)

    # WHEN getting the run parameters file name
    run_parameters_path: Path = flow_cell.run_parameters_path

    # THEN it should exist and be the expected one
    assert run_parameters_path.exists()
    assert run_parameters_path.name == expected_run_parameters_file_name


def test_run_parameters_path_when_non_existing(tmp_flow_cells_directory_no_run_parameters: Path):
    """Test that getting the path of the run parameters path fails if the file does not exist."""
    # GIVEN a flowcell object with a directory without a run parameters file
    flow_cell = FlowCellDirectoryData(flow_cell_path=tmp_flow_cells_directory_no_run_parameters)

    # WHEN fetching the run parameters path
    with pytest.raises(FlowCellError) as exc:
        # THEN a FlowCellError is raised
        flow_cell.run_parameters_path
    assert "No run parameters file found in flow cell" in str(exc.value)


@pytest.mark.parametrize(
    "flow_cell_fixture, expected_sequencer",
    [
        ("hiseq_2500_flow_cell", Sequencers.HISEQGA),
        ("hiseq_x_flow_cell", Sequencers.HISEQX),
        ("novaseq_6000_flow_cell", Sequencers.NOVASEQ),
        ("novaseq_x_flow_cell", Sequencers.NOVASEQX),
    ],
)
def test_flow_cell_run_parameters_type(
    flow_cell_fixture: str, expected_sequencer: str, request: FixtureRequest
):
    """Test that the run parameters of the flow cell is of the expected type."""
    # GIVEN a flow cell without _run_parameters
    flow_cell: FlowCellDirectoryData = request.getfixturevalue(flow_cell_fixture)
    assert not flow_cell._run_parameters

    # WHEN creating the run parameters of the flow cell
    run_parameters: RunParameters = flow_cell.run_parameters

    # THEN the run parameters sequencer is the same as of the flow cell
    assert run_parameters.sequencer == expected_sequencer


def test_has_demultiplexing_started_locally_false(tmp_flow_cell_directory_bclconvert: Path):
    # GIVEN a flow cell without a demuxstarted.txt file
    flow_cell = FlowCellDirectoryData(tmp_flow_cell_directory_bclconvert)
    assert not Path(flow_cell.path, DemultiplexingDirsAndFiles.DEMUX_STARTED).exists()

    # WHEN checking if the flow cell has started demultiplexing
    has_demux_started: bool = flow_cell.has_demultiplexing_started_locally()

    # THEN the response should be False
    assert not has_demux_started


def test_has_demultiplexing_started_locally_true(
    tmp_flow_cell_directory_bclconvert: Path,
):
    # GIVEN a flow cell with a demuxstarted.txt file
    flow_cell = FlowCellDirectoryData(tmp_flow_cell_directory_bclconvert)
    Path(flow_cell.path, DemultiplexingDirsAndFiles.DEMUX_STARTED).touch()

    # WHEN checking if the flow cell has started demultiplexing
    has_demux_started: bool = flow_cell.has_demultiplexing_started_locally()

    # THEN the response should be True
    assert has_demux_started


def test_has_demultiplexing_started_on_sequencer_true(
    novaseqx_flow_cell_dir_with_analysis_data: Path,
):
    # GIVEN a flow cell with a BCLConvert folder
    flow_cell = FlowCellDirectoryData(novaseqx_flow_cell_dir_with_analysis_data)
    Path.mkdir(
        Path(
            flow_cell.path,
            get_latest_analysis_path(flow_cell.path),
            DemultiplexingDirsAndFiles.DATA,
            DemultiplexingDirsAndFiles.BCL_CONVERT,
        )
    )

    # WHEN checking if the flow cell has started demultiplexing
    has_demux_started: bool = flow_cell.has_demultiplexing_started_on_sequencer()

    # THEN the response should be True
    assert has_demux_started


def test_has_demultiplexing_started_on_sequencer_false(
    novaseqx_flow_cell_dir_with_analysis_data: Path,
):
    # GIVEN a flow cell without a BCLConvert folder
    flow_cell = FlowCellDirectoryData(novaseqx_flow_cell_dir_with_analysis_data)
    assert not Path(
        flow_cell.path,
        get_latest_analysis_path(flow_cell.path),
        DemultiplexingDirsAndFiles.DATA,
        DemultiplexingDirsAndFiles.BCL_CONVERT,
    ).exists()

    # WHEN checking if the flow cell has started demultiplexing
    has_demux_started: bool = flow_cell.has_demultiplexing_started_on_sequencer()

    # THEN the response should be False
    assert not has_demux_started



"""Test module cg.cli.upload.nipt.ftp"""

import logging

from click.testing import CliRunner

from cg.cli.upload.nipt.ftp import nipt_upload_all, nipt_upload_case
from cg.meta.upload.nipt import NiptUploadAPI
from cg.models.cg_config import CGConfig


def test_nipt_upload_case(upload_context: CGConfig, cli_runner: CliRunner, caplog, mocker):
    """Tests CLI command to upload a single case"""

    caplog.set_level(logging.DEBUG)
    # GIVEN a specified NIPT case
    case_id = "angrybird"

    # WHEN adding a result file of a specified NIPT case
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    mocker.patch.object(NiptUploadAPI, "flowcell_passed_qc_value")
    result = cli_runner.invoke(nipt_upload_case, [case_id], obj=upload_context)

    # THEN the nipt upload should start and exit without errors
    assert "*** NIPT FTP UPLOAD START ***" in caplog.text
    assert result.exit_code == 0


def test_nipt_upload_case_dry(upload_context: CGConfig, cli_runner: CliRunner, caplog, mocker):
    """Tests CLI command to upload a single case"""

    caplog.set_level(logging.DEBUG)
    # GIVEN a specified NIPT case
    case_id = "angrybird"

    # WHEN adding a result file of a specified NIPT case (dry run)
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    mocker.patch.object(NiptUploadAPI, "flowcell_passed_qc_value")
    result = cli_runner.invoke(nipt_upload_case, ["--dry-run", case_id], obj=upload_context)

    # THEN the nipt upload should start and exit without errors
    assert "*** NIPT FTP UPLOAD START ***" in caplog.text
    assert result.exit_code == 0


def test_nipt_upload_all(upload_context: CGConfig, cli_runner: CliRunner, caplog, mocker):
    """Tests CLI command to upload a single case"""

    caplog.set_level(logging.DEBUG)
    # GIVEN adding all available NIPT case results

    # WHEN adding a result file of a all available NIPT cases
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    mocker.patch.object(NiptUploadAPI, "flowcell_passed_qc_value")
    result = cli_runner.invoke(nipt_upload_all, obj=upload_context)

    # THEN the NIPT upload should start and exit without errors
    assert "*** UPLOAD ALL AVAILABLE NIPT RESULTS ***" in caplog.text
    assert result.exit_code == 0


def test_nipt_upload_all_dry(upload_context: CGConfig, cli_runner: CliRunner, caplog, mocker):
    """Tests CLI command to upload a single case"""

    caplog.set_level(logging.DEBUG)
    # GIVEN adding all available NIPT case results

    # WHEN adding a result file of a all available NIPT cases (dry run)
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    mocker.patch.object(NiptUploadAPI, "flowcell_passed_qc_value")
    result = cli_runner.invoke(nipt_upload_all, ["--dry-run"], obj=upload_context)

    # THEN the NIPT upload should start and exit without errors
    assert "*** UPLOAD ALL AVAILABLE NIPT RESULTS ***" in caplog.text
    assert result.exit_code == 0


def test_nipt_upload_case_not_changing_uploaded_at(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a single case"""

    # GIVEN a specified NIPT case that has its analysis stored but is not yet uploaded
    caplog.set_level(logging.DEBUG)

    analysis_obj = helpers.add_analysis(store=upload_context.status_db)
    case_id = analysis_obj.case.internal_id
    assert not analysis_obj.upload_started_at
    assert not analysis_obj.uploaded_at

    # WHEN adding a result file of a specified NIPT case
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    mocker.patch.object(NiptUploadAPI, "flowcell_passed_qc_value")
    result = cli_runner.invoke(nipt_upload_case, [case_id], obj=upload_context)

    # THEN set analysis.upload_started_at in the database
    assert not analysis_obj.upload_started_at

    # THEN set analysis.uploaded_at in the database
    assert not analysis_obj.uploaded_at

    # THEN exit without errors
    assert result.exit_code == 0



"""Delivery report constants."""
from cg.constants import DataDelivery
from cg.constants.constants import Pipeline

BALSAMIC_REPORT_ACCREDITED_PANELS: list[str] = ["gmsmyeloid"]

REPORT_SUPPORTED_PIPELINES: tuple[Pipeline, ...] = (
    Pipeline.BALSAMIC,
    Pipeline.BALSAMIC_UMI,
    Pipeline.BALSAMIC_QC,
    Pipeline.MIP_DNA,
    Pipeline.RNAFUSION,
)

REPORT_SUPPORTED_DATA_DELIVERY: tuple[DataDelivery, ...] = (
    DataDelivery.ANALYSIS_FILES,
    DataDelivery.ANALYSIS_SCOUT,
    DataDelivery.FASTQ_ANALYSIS,
    DataDelivery.FASTQ_ANALYSIS_SCOUT,
    DataDelivery.FASTQ_QC_ANALYSIS,
    DataDelivery.FASTQ_SCOUT,
    DataDelivery.SCOUT,
)

NA_FIELD: str = "N/A"
YES_FIELD: str = "Ja"
NO_FIELD: str = "Nej"
PRECISION: int = 2

REPORT_GENDER: dict[str, str] = {
    "unknown": "Okänd",
    "female": "Kvinna",
    "male": "Man",
}

BALSAMIC_ANALYSIS_TYPE: dict[str, str] = {
    "tumor_wgs": "Tumör-endast (helgenomsekvensering)",
    "tumor_normal_wgs": "Tumör/normal (helgenomsekvensering)",
    "tumor_panel": "Tumör-endast (panelsekvensering)",
    "tumor_normal_panel": "Tumör/normal (panelsekvensering)",
}

# Report required fields (OPTIONAL: "version")
REQUIRED_REPORT_FIELDS: list[str] = [
    "customer",
    "date",
    "case",
    "accredited",
]

# Customer required fields (OPTIONAL: "id")
REQUIRED_CUSTOMER_FIELDS: list[str] = [
    "name",
    "invoice_address",
    "scout_access",
]

# Case required fields
REQUIRED_CASE_FIELDS: list[str] = [
    "name",
    "id",
    "samples",
    "data_analysis",
    "applications",
]

# Application required fields (OPTIONAL: "version", "prep_category", "description", "limitations", "external")
REQUIRED_APPLICATION_FIELDS: list[str] = [
    "tag",
    "accredited",
]

# Data analysis required fields
REQUIRED_DATA_ANALYSIS_FIELDS: list[str] = [
    "customer_pipeline",
    "pipeline",
    "pipeline_version",
    "genome_build",
]

REQUIRED_DATA_ANALYSIS_MIP_DNA_FIELDS: list[str] = REQUIRED_DATA_ANALYSIS_FIELDS + [
    "panels",
]

REQUIRED_DATA_ANALYSIS_BALSAMIC_FIELDS: list[str] = REQUIRED_DATA_ANALYSIS_FIELDS + [
    "type",
    "variant_callers",
]

REQUIRED_DATA_ANALYSIS_RNAFUSION_FIELDS: list[str] = REQUIRED_DATA_ANALYSIS_FIELDS

# Sample required fields
_REQUIRED_SAMPLE_FIELDS: list[str] = [
    "name",
    "id",
    "ticket",
    "gender",
    "source",
    "application",
    "methods",
    "metadata",
    "timestamps",
]

REQUIRED_SAMPLE_MIP_DNA_FIELDS: list[str] = _REQUIRED_SAMPLE_FIELDS + [
    "status",
]

REQUIRED_SAMPLE_BALSAMIC_FIELDS: list[str] = _REQUIRED_SAMPLE_FIELDS + [
    "tumour",
]

REQUIRED_SAMPLE_RNAFUSION_FIELDS: list[str] = REQUIRED_SAMPLE_BALSAMIC_FIELDS

# Methods required fields (OPTIONAL: "library_prep", "sequencing")
REQUIRED_SAMPLE_METHODS_FIELDS: list[str] = []

# Timestamp required fields (OPTIONAL: "prepared_at", "reads_updated_at")
REQUIRED_SAMPLE_TIMESTAMP_FIELDS: list[str] = [
    "ordered_at",
    "received_at",  # Optional for external samples
]

# Metadata required fields
_REQUIRED_SAMPLE_METADATA_FIELDS: list[str] = [
    "million_read_pairs",
    "duplicates",
]

REQUIRED_SAMPLE_METADATA_MIP_DNA_WGS_FIELDS: list[str] = _REQUIRED_SAMPLE_METADATA_FIELDS + [
    "gender",
    "mapped_reads",
    "mean_target_coverage",
    "pct_10x",
]

REQUIRED_SAMPLE_METADATA_MIP_DNA_FIELDS: list[str] = REQUIRED_SAMPLE_METADATA_MIP_DNA_WGS_FIELDS + [
    "bait_set",
]

_REQUIRED_SAMPLE_METADATA_BALSAMIC_FIELDS: list[str] = _REQUIRED_SAMPLE_METADATA_FIELDS + [
    "mean_insert_size",
    "fold_80",
]

REQUIRED_SAMPLE_METADATA_BALSAMIC_TARGETED_FIELDS: list[
    str
] = _REQUIRED_SAMPLE_METADATA_BALSAMIC_FIELDS + [
    "bait_set",
    "bait_set_version",
    "median_target_coverage",
    "pct_250x",
    "pct_500x",
    "gc_dropout",
]

REQUIRED_SAMPLE_METADATA_BALSAMIC_TO_WGS_FIELDS: list[
    str
] = _REQUIRED_SAMPLE_METADATA_BALSAMIC_FIELDS + [
    "median_coverage",
    "pct_60x",
    "pct_reads_improper_pairs",
]

REQUIRED_SAMPLE_METADATA_BALSAMIC_TN_WGS_FIELDS: list[
    str
] = REQUIRED_SAMPLE_METADATA_BALSAMIC_TO_WGS_FIELDS + [
    "pct_15x",
]

REQUIRED_SAMPLE_METADATA_RNAFUSION_FIELDS: list[str] = _REQUIRED_SAMPLE_METADATA_FIELDS + [
    "bias_5_3",
    "gc_content",
    "input_amount",
    "mapped_reads",
    "mean_length_r1",
    "mrna_bases",
    "pct_adapter",
    "pct_surviving",
    "q20_rate",
    "q30_rate",
    "ribosomal_bases",
    "rin",
    "uniquely_mapped_reads",
]



import logging
import os.path
from pathlib import Path
from tempfile import TemporaryDirectory

import requests
from flask import Flask

from cg.constants.constants import FileFormat
from cg.exc import TicketCreationError
from cg.io.controller import WriteFile, WriteStream

LOG = logging.getLogger(__name__)
TEXT_FILE_ATTACH_PARAMS = "data:text/plain;charset=utf-8,{content}"


class OsTicket(object):

    """Interface to ticket system"""

    def __init__(self):
        self.headers = None
        self.url = None
        self.osticket_email = None
        self.email_uri = None

    def init_app(self, app: Flask):
        """Initialize the API in Flask"""
        self.setup(
            api_key=app.config["OSTICKET_API_KEY"],
            domain=app.config["OSTICKET_DOMAIN"],
            osticket_email=app.config["SUPPORT_SYSTEM_EMAIL"],
            email_uri=app.config["EMAIL_URI"],
        )

    def setup(
        self,
        api_key: str = None,
        domain: str = None,
        osticket_email: str = None,
        email_uri: str = None,
    ):
        """Initialize the API"""
        self.headers = {"X-API-Key": api_key}
        self.url = os.path.join(domain, "api/tickets.json")
        self.osticket_email = osticket_email
        self.email_uri = email_uri

    def open_ticket(
        self, attachment: dict, email: str, message: str, name: str, subject: str
    ) -> str:
        """Open a new ticket through the REST API"""
        data = dict(
            name=name,
            email=email,
            subject=subject,
            message=message,
            attachments=[attachment],
        )
        res = requests.post(self.url, json=data, headers=self.headers)
        if res.ok:
            return res.text
        LOG.error(f"res.text: {res.text}, reason: {res.reason}")
        raise TicketCreationError(res)

    @staticmethod
    def create_new_ticket_attachment(content: dict, file_name: str) -> dict:
        return {
            file_name: TEXT_FILE_ATTACH_PARAMS.format(
                content=WriteStream.write_stream_from_content(
                    content=content, file_format=FileFormat.JSON
                )
            )
        }

    @staticmethod
    def create_connecting_ticket_attachment(content: dict) -> TemporaryDirectory:
        directory = TemporaryDirectory()
        WriteFile.write_file_from_content(
            content=content,
            file_format=FileFormat.JSON,
            file_path=Path(directory.name, "order.json"),
        )
        return directory



"""Fixtures for cli workflow raredisease tests"""

import pytest

from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.cg_config import CGConfig


@pytest.fixture(scope="function", name="raredisease_context")
def fixture_raredisease_context(
    cg_context: CGConfig,
) -> CGConfig:
    """Context to use in cli."""
    cg_context.meta_apis["analysis_api"] = RarediseaseAnalysisAPI(config=cg_context)
    return cg_context



"""Code that handles uploading to mutacc from the CLI"""
import logging

import click

from cg.apps.mutacc_auto import MutaccAutoAPI
from cg.apps.scout.scout_export import ScoutExportCase
from cg.apps.scout.scoutapi import ScoutAPI
from cg.meta.upload.mutacc import UploadToMutaccAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.command("process-solved")
@click.option("-c", "--case-id", help="internal case id, leave empty to process all")
@click.option("-d", "--days-ago", type=int, default=1, help="days since solved")
@click.option("-C", "--customers", type=str, multiple=True, help="Filter on customers")
@click.option("--dry-run", is_flag=True, help="only print cases to be processed")
@click.pass_obj
def process_solved(
    context: CGConfig, case_id: str | None, days_ago: int, customers: tuple[str], dry_run: bool
):
    """Process cases with mutacc that has been marked as solved in scout.
    This prepares them to be uploaded to the mutacc database"""

    LOG.info("----------------- PROCESS-SOLVED ----------------")

    scout_api: ScoutAPI = context.scout_api
    mutacc_auto_api: MutaccAutoAPI = context.mutacc_auto_api
    mutacc_upload_api = UploadToMutaccAPI(scout_api=scout_api, mutacc_auto_api=mutacc_auto_api)

    # Get cases to upload into mutacc from scout
    finished_cases: list[ScoutExportCase] = []
    if case_id is not None:
        finished_cases = scout_api.get_cases(finished=True, case_id=case_id)
    elif days_ago is not None:
        finished_cases = scout_api.get_solved_cases(days_ago=days_ago)
    else:
        LOG.info("Please enter option '--case-id' or '--days-ago'")

    number_processed = 0
    for case in finished_cases:
        number_processed += 1
        if customers and case.owner not in customers:
            LOG.info(f"skipping {case.id}: Not valid customer {case.owner}")
            continue
        if dry_run:
            LOG.info(f"Would process case {case.id} with mutacc")
            continue

        LOG.info(f"Start processing case {case.id} with mutacc")
        mutacc_upload_api.extract_reads(case)

    if number_processed == 0:
        LOG.info(f"No cases were solved within the last {days_ago} days")


@click.command("processed-solved")
@click.pass_obj
def processed_solved(context: CGConfig):
    """Upload solved cases that has been processed by mutacc to the mutacc database"""

    mutacc_auto_api: MutaccAutoAPI = context.mutacc_auto_api

    LOG.info("----------------- PROCESSED-SOLVED ----------------")

    LOG.info("Uploading processed cases by mutacc to the mutacc database")

    mutacc_auto_api.import_reads()



"""Delivery report upload to scout commands."""
import logging

import click
from housekeeper.store.models import Version

from cg.cli.generate.report.options import ARGUMENT_CASE_ID
from cg.cli.generate.report.utils import get_report_api, get_report_case
from cg.meta.report.report_api import ReportAPI
from cg.store.models import Case

LOG = logging.getLogger(__name__)


@click.command("delivery-report-to-scout")
@ARGUMENT_CASE_ID
@click.option(
    "-r", "--re-upload", is_flag=True, default=False, help="Re-upload existing delivery report"
)
@click.option(
    "-d", "--dry-run", is_flag=True, default=False, help="Run command without uploading to Scout"
)
@click.pass_context
def upload_delivery_report_to_scout(
    context: click.Context, case_id: str, re_upload: bool, dry_run: bool
) -> None:
    """Fetches a delivery report from Housekeeper and uploads it to Scout."""
    click.echo(click.style("--------------- DELIVERY REPORT UPLOAD ---------------"))
    case: Case = get_report_case(context, case_id)
    report_api: ReportAPI = get_report_api(context, case)
    version: Version = report_api.housekeeper_api.last_version(case_id)
    delivery_report: str | None = report_api.get_delivery_report_from_hk(
        case_id=case_id, version=version
    )
    if delivery_report and not dry_run:
        report_api.scout_api.upload_delivery_report(
            report_path=delivery_report, case_id=case.internal_id, update=re_upload
        )
        return
    LOG.error("Delivery report not uploaded to Scout")



import logging
from pathlib import Path

import pytest
from _pytest.fixtures import FixtureRequest

from cg.constants.demultiplexing import RunParametersXMLNodes
from cg.constants.sequencing import Sequencers
from cg.exc import RunParametersError, XMLError
from cg.models.demultiplex.run_parameters import (
    RunParameters,
    RunParametersHiSeq,
    RunParametersNovaSeq6000,
    RunParametersNovaSeqX,
)


@pytest.mark.parametrize(
    "run_parameters_path_fixture",
    [
        "hiseq_x_single_index_run_parameters_path",
        "hiseq_2500_double_index_run_parameters_path",
        "novaseq_6000_run_parameters_path",
        "novaseq_x_run_parameters_path",
    ],
)
def test_run_parameters_parent_class_fails(
    run_parameters_path_fixture: str, request: FixtureRequest
):
    """Test that trying to instantiate the parent RunParameters class raises an error."""
    # GIVEN a valid path for a run parameters file
    run_parameters_path: Path = request.getfixturevalue(run_parameters_path_fixture)

    # WHEN trying to instantiate the parent class with a RunParameters path
    with pytest.raises(NotImplementedError) as exc_info:
        # THEN an NotImplementedError is raised
        RunParameters(run_parameters_path=run_parameters_path)
    assert "Parent class instantiated" in str(exc_info)


def test_run_parameters_hiseq_x(hiseq_x_single_index_run_parameters_path: Path):
    """Tests that creating a HiSeq RunParameters with a HiSeqX path file works."""
    # GIVEN a valid HiSeqX run parameters file path

    # WHEN creating a HiSeqX RunParameters object
    run_parameters = RunParametersHiSeq(hiseq_x_single_index_run_parameters_path)

    # THEN the created object is of the correct type
    assert isinstance(run_parameters, RunParametersHiSeq)
    # THEN the sequencer is HiSeqX
    assert run_parameters.sequencer == Sequencers.HISEQX


def test_run_parameters_hiseq_2500(hiseq_2500_double_index_run_parameters_path: Path):
    """Tests that creating a HiSeq RunParameters with a HiSeq2500 path file works."""
    # GIVEN a valid HiSeq2500 run parameters file path

    # WHEN creating a HiSeq2500 RunParameters object
    run_parameters = RunParametersHiSeq(hiseq_2500_double_index_run_parameters_path)

    # THEN the created object is of the correct type
    assert isinstance(run_parameters, RunParametersHiSeq)
    # THEN the sequencer is HiSeq2500
    assert run_parameters.sequencer == Sequencers.HISEQGA


def test_run_parameters_hiseq_wrong_file(novaseq_6000_run_parameters_path: Path):
    """Tests that creating a RunParameters HiSeq object with the wrong file fails."""
    # GIVEN a file path with a run parameters file from an instrument different from HiSeq

    # WHEN trying to create a HiSeq RunParameters object with the file
    with pytest.raises(RunParametersError) as exc_info:
        # THEN a RunParametersError is raised
        RunParametersHiSeq(run_parameters_path=novaseq_6000_run_parameters_path)
    assert (
        f"Could not find node {RunParametersXMLNodes.APPLICATION_NAME} in the run parameters file"
        in str(exc_info.value)
    )


def test_run_parameters_novaseq_6000(novaseq_6000_run_parameters_path: Path):
    """Tests that creating a RunParameters NovaSeq6000 object with the correct file works."""
    # GIVEN a valid NovaSeq6000 run parameters file path

    # WHEN creating a NovaSeq6000 RunParameters object
    run_parameters = RunParametersNovaSeq6000(run_parameters_path=novaseq_6000_run_parameters_path)

    # THEN the created object is of the correct type and has the correct attributes
    assert isinstance(run_parameters, RunParametersNovaSeq6000)
    assert run_parameters.sequencer == Sequencers.NOVASEQ


def test_run_parameters_novaseq_6000_wrong_file(novaseq_x_run_parameters_path: Path):
    """Tests that creating a RunParameters NovaSeq6000 object with the wrong file fails."""
    # GIVEN a file path with a run parameters file from an instrument different from NovaSeq6000

    # WHEN trying to create a NovaSeq6000 RunParameters object with the file
    with pytest.raises(RunParametersError) as exc_info:
        # THEN a RunParametersError is raised
        RunParametersNovaSeq6000(run_parameters_path=novaseq_x_run_parameters_path)
    assert (
        str(exc_info.value)
        == f"The file parsed does not correspond to {RunParametersXMLNodes.NOVASEQ_6000_APPLICATION}"
    )


def test_run_parameters_novaseq_x(novaseq_x_run_parameters_path: Path):
    """Tests that creating a RunParameters NovaSeqX object with the correct file works."""
    # GIVEN a valid NovaSeqX run parameters file path

    # WHEN creating a NovaSeqX RunParameters object
    run_parameters = RunParametersNovaSeqX(run_parameters_path=novaseq_x_run_parameters_path)

    # THEN the created object is of the correct type and has the correct attributes
    assert isinstance(run_parameters, RunParametersNovaSeqX)
    assert run_parameters.sequencer == Sequencers.NOVASEQX


def test_run_parameters_novaseq_x_wrong_file(novaseq_6000_run_parameters_path: Path):
    """Tests that creating a RunParameters NovaSeqX object with the wrong file fails."""
    # GIVEN a file path with a run parameters file from an instrument different from NovaSeqX

    # WHEN trying to create a NovaSeqX RunParameters object with the file
    with pytest.raises(RunParametersError) as exc_info:
        # THEN a RunParametersError is raised
        RunParametersNovaSeqX(run_parameters_path=novaseq_6000_run_parameters_path)
    assert (
        f"Could not find node {RunParametersXMLNodes.INSTRUMENT_TYPE} in the run parameters"
        in str(exc_info.value)
    )


def test_run_parameters_novaseq_x_file_wrong_instrument(run_parameters_wrong_instrument: Path):
    """Test that initialising a NovaSeqX with a RunParameters file with a wrong instrument fails."""
    # GIVEN a RunParameters file with a wrong instrument

    # WHEN initialising the NovaSeqX
    with pytest.raises(RunParametersError) as exc_info:
        # THEN assert that an exception was raised since the control software version was not found
        RunParametersNovaSeqX(run_parameters_path=run_parameters_wrong_instrument)

    assert (
        str(exc_info.value)
        == f"The file parsed does not correspond to {RunParametersXMLNodes.NOVASEQ_X_INSTRUMENT}"
    )


@pytest.mark.parametrize(
    "run_parameters_fixture",
    [
        "hiseq_2500_run_parameters_double_index",
        "hiseq_x_run_parameters_single_index",
        "novaseq_x_run_parameters",
    ],
)
def test_reagent_kit_version_hiseq_and_novaseq_x(
    run_parameters_fixture: str, request: FixtureRequest
):
    """Test that getting reagent kit version from a HiSeq or NovaSeqX RunParameters returns None."""
    # GIVEN a valid RunParameters object
    run_parameters: RunParameters = request.getfixturevalue(run_parameters_fixture)

    # WHEN fetching the reagent kit version

    # THEN the reagent kit version is None
    assert not run_parameters.reagent_kit_version


def test_reagent_kit_version_novaseq_6000(novaseq_6000_run_parameters: RunParametersNovaSeq6000):
    """Test that getting reagent kit version from a correct file returns an expected value."""
    # GIVEN a valid RunParameters object for NovaSeq6000

    # WHEN fetching the reagent kit version
    reagent_kit_version: str = novaseq_6000_run_parameters.reagent_kit_version

    # THEN the reagent kit version exists and is not "unknown"
    assert reagent_kit_version
    assert reagent_kit_version != RunParametersXMLNodes.UNKNOWN_REAGENT_KIT_VERSION


def test_reagent_kit_version_novaseq_6000_missing_version(
    run_parameters_missing_versions: RunParametersNovaSeq6000, caplog
):
    """Test that 'unknown' will be returned if the run parameters file has no reagent kit method."""
    caplog.set_level(logging.INFO)
    # GIVEN a RunParameters object for NovaSeq6000 created from a file without reagent kit version

    # WHEN fetching the reagent kit version
    reagent_kit: str = run_parameters_missing_versions.reagent_kit_version

    # THEN the reagent kit version is unknown
    assert reagent_kit == RunParametersXMLNodes.UNKNOWN_REAGENT_KIT_VERSION
    assert "Could not determine reagent kit version" in caplog.text


@pytest.mark.parametrize(
    "run_parameters_fixture",
    [
        "hiseq_2500_run_parameters_double_index",
        "hiseq_x_run_parameters_single_index",
        "novaseq_x_run_parameters",
    ],
)
def test_control_software_version_hiseq_and_novaseq_x(
    run_parameters_fixture: str, request: FixtureRequest
):
    """Test that getting control software version from HiSeq/NovaSeqX RunParameters returns None."""
    # GIVEN a valid RunParameters object
    run_parameters: RunParameters = request.getfixturevalue(run_parameters_fixture)

    # WHEN fetching the control software version

    # THEN the control software version is None
    assert not run_parameters.control_software_version


def test_control_software_version_novaseq_6000(
    novaseq_6000_run_parameters: RunParametersNovaSeq6000,
):
    """Test that getting control software version from a correct file returns an expected value."""
    # GIVEN a valid RunParameters object for NovaSeq6000

    # WHEN fetching the control software version
    control_software_version: str = novaseq_6000_run_parameters.control_software_version

    # THEN the control software version is a non-empty string
    assert isinstance(control_software_version, str)
    assert control_software_version != ""


def test_control_software_version_novaseq_6000_no_version(
    run_parameters_missing_versions: Path, caplog
):
    """Test that fetching the control software version from a file without that field fails."""
    caplog.set_level(logging.INFO)
    # GIVEN a RunParameters object created from a file without control software version

    # WHEN fetching the control software version
    with pytest.raises(XMLError):
        # THEN assert that an exception was raised since the control software version was not found
        run_parameters_missing_versions.control_software_version

    assert (
        f"Could not find node with name {RunParametersXMLNodes.APPLICATION_VERSION} in XML tree"
        in caplog.text
    )


@pytest.mark.parametrize(
    "run_parameters_fixture, expected_index_length",
    [
        ("hiseq_2500_run_parameters_double_index", 8),
        ("hiseq_x_run_parameters_single_index", 8),
        ("novaseq_6000_run_parameters", 10),
        ("novaseq_x_run_parameters", 10),
    ],
)
def test_index_length(
    run_parameters_fixture: str, expected_index_length: int, request: FixtureRequest
):
    """Test getting the index length from RunParameters objects return expected values."""
    # GIVEN a valid RunParameters object
    run_parameters: RunParameters = request.getfixturevalue(run_parameters_fixture)

    # WHEN getting the index length
    real_index_length: int = run_parameters.index_length

    # THEN the index length is an int
    assert isinstance(real_index_length, int)
    # THEN the index length is the expected value
    assert real_index_length == expected_index_length


@pytest.mark.parametrize(
    "run_parameters_fixture",
    [
        "run_parameters_novaseq_6000_different_index",
        "run_parameters_novaseq_x_different_index",
    ],
)
def test_index_length_different_length(
    run_parameters_fixture: str,
    request: FixtureRequest,
):
    """Test that getting the index length from a file with different index cycles fails."""
    # GIVEN a NovaSeq6000 RunParameters object created from a file with different index cycles
    run_parameters: RunParameters = request.getfixturevalue(run_parameters_fixture)

    # WHEN fetching index length
    with pytest.raises(RunParametersError) as exc_info:
        # THEN assert that an exception was raised since the index cycles are different
        run_parameters.index_length
    assert str(exc_info.value) == "Index lengths are not the same!"


@pytest.mark.parametrize(
    "run_parameters_fixture",
    [
        "hiseq_2500_run_parameters_double_index",
        "hiseq_x_run_parameters_single_index",
        "novaseq_6000_run_parameters",
        "novaseq_x_run_parameters",
    ],
)
def test_get_cycles(run_parameters_fixture: str, request: FixtureRequest):
    """Test that the read and index cycles are read correctly for any RunParameters object."""
    # GIVEN a RunParameters object
    run_parameters: RunParameters = request.getfixturevalue(run_parameters_fixture)

    # WHEN getting any read cycle
    read_cycles: list[int] = [
        run_parameters.get_read_1_cycles(),
        run_parameters.get_read_2_cycles(),
        run_parameters.get_index_1_cycles(),
        run_parameters.get_index_2_cycles(),
    ]

    # THEN all read cycles are non-negative integers
    for cycles in read_cycles:
        assert isinstance(cycles, int)
        assert cycles >= 0



"""Tests for MicroSALT analysis."""
import logging
from pathlib import Path

import mock

from cg.constants.constants import CaseActions, Pipeline
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.orders.sample_base import ControlEnum
from cg.store import Store
from cg.store.models import Case
from tests.mocks.tb_mock import MockTB
from tests.store_helpers import StoreHelpers


def test_qc_check_fail(
    qc_microsalt_context: CGConfig,
    microsalt_qc_fail_run_dir_path: Path,
    microsalt_qc_fail_lims_project: str,
    microsalt_case_qc_fail: str,
    caplog,
    mocker,
):
    """QC check for a microsalt case that should fail."""
    caplog.set_level(logging.INFO)
    store: Store = qc_microsalt_context.status_db
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]

    # GIVEN a case that is to be stored
    microsalt_case: Case = store.get_case_by_internal_id(internal_id=microsalt_case_qc_fail)
    for index in range(4):
        microsalt_case.samples[index].reads = 1000

    mocker.patch.object(MicrosaltAnalysisAPI, "create_qc_done_file")

    # WHEN performing QC check
    qc_pass: bool = microsalt_api.microsalt_qc(
        case_id=microsalt_case_qc_fail,
        run_dir_path=microsalt_qc_fail_run_dir_path,
        lims_project=microsalt_qc_fail_lims_project,
    )

    # THEN the QC should fail
    assert not qc_pass
    assert "failed" in caplog.text


def test_qc_check_pass(
    qc_microsalt_context: CGConfig,
    microsalt_qc_pass_run_dir_path: Path,
    microsalt_qc_pass_lims_project: str,
    microsalt_case_qc_pass: str,
    caplog,
    mocker,
):
    """QC check for a microsalt case that should pass."""
    caplog.set_level(logging.INFO)
    store: Store = qc_microsalt_context.status_db
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]

    # GIVEN a case that is to be stored
    microsalt_case: Case = store.get_case_by_internal_id(internal_id=microsalt_case_qc_pass)
    microsalt_case.samples[1].control = ControlEnum.negative
    microsalt_case.samples[1].reads = 1100000

    mocker.patch.object(MicrosaltAnalysisAPI, "create_qc_done_file")

    # WHEN performing QC check
    qc_pass: bool = microsalt_api.microsalt_qc(
        case_id=microsalt_case_qc_pass,
        run_dir_path=microsalt_qc_pass_run_dir_path,
        lims_project=microsalt_qc_pass_lims_project,
    )

    # THEN the QC should pass
    assert qc_pass
    assert "passed" in caplog.text


def test_qc_check_negative_control_fail(
    qc_microsalt_context: CGConfig,
    microsalt_qc_fail_run_dir_path: Path,
    microsalt_qc_fail_lims_project: str,
    microsalt_case_qc_fail: str,
    caplog,
    mocker,
):
    """QC check for a microsalt case where a negative control fails QC."""

    caplog.set_level(logging.INFO)
    store = qc_microsalt_context.status_db
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]

    # GIVEN a case that is to be stored
    microsalt_case: Case = store.get_case_by_internal_id(internal_id=microsalt_case_qc_fail)
    microsalt_case.samples[0].control = ControlEnum.negative

    mocker.patch.object(MicrosaltAnalysisAPI, "create_qc_done_file")

    # WHEN performing QC check
    qc_pass: bool = microsalt_api.microsalt_qc(
        case_id=microsalt_case_qc_fail,
        run_dir_path=microsalt_qc_fail_run_dir_path,
        lims_project=microsalt_qc_fail_lims_project,
    )

    # THEN the QC should fail
    assert not qc_pass
    assert "failed" in caplog.text
    assert "Negative control sample" in caplog.text


def test_get_latest_case_path(
    mocker,
    qc_microsalt_context: CGConfig,
    microsalt_case_qc_pass: str,
    microsalt_analysis_dir: Path,
):
    """Test get_latest_case_path return the first case path and not single sample path"""
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]

    # GIVEN a case with different case paths, both single sample and case analyses
    mocker.patch.object(MicrosaltAnalysisAPI, "get_project", return_value="ACC12345")
    mocker.patch.object(
        MicrosaltAnalysisAPI,
        "get_case_path",
        return_value=[
            Path(microsalt_analysis_dir, "ACC12345A2_2023"),
            Path(microsalt_analysis_dir, "ACC12345_2022"),
            Path(microsalt_analysis_dir, "ACC12345A1_2023"),
        ],
    )
    # WHEN getting the latest case path
    path = microsalt_api.get_latest_case_path(case_id=microsalt_case_qc_pass)

    # THEN the first case path should be returned
    assert Path(microsalt_analysis_dir, "ACC12345_2022") == path


def test_get_cases_to_store(
    qc_microsalt_context: CGConfig, helpers: StoreHelpers, trailblazer_api: MockTB
):
    """Test that the cases fetched are Microsalt and finished successfully."""
    # GIVEN a MicrosaltAPI, a Store and a TrailblazerAPI
    analysis_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]
    store: Store = analysis_api.status_db
    mock.patch.object(trailblazer_api, "is_latest_analysis_completed", return_value=True)
    analysis_api.trailblazer_api = trailblazer_api

    # GIVEN a running case in the store
    helpers.ensure_case(store=store, data_analysis=Pipeline.MICROSALT, action=CaseActions.RUNNING)

    # WHEN getting the cases to store in Housekeeper
    cases_to_store: list[Case] = analysis_api.get_cases_to_store()
    case: Case = cases_to_store[0]

    # THEN a list with one microsalt case is returned
    assert len(cases_to_store) == 1
    assert case.data_analysis == Pipeline.MICROSALT
    assert case.action == CaseActions.RUNNING



"""Module for parsing sequencing run metadata from RunParameters file."""
import logging
from abc import abstractmethod
from pathlib import Path
from xml.etree import ElementTree

from cg.constants.demultiplexing import RunParametersXMLNodes
from cg.constants.sequencing import SEQUENCER_TYPES, Sequencers
from cg.exc import RunParametersError, XMLError
from cg.io.xml import get_tree_node, read_xml

LOG = logging.getLogger(__name__)


class RunParameters:
    """Base class with basic functions to handle the run parameters from a sequencing run."""

    def __init__(self, run_parameters_path: Path):
        self.path: Path = run_parameters_path
        self.tree: ElementTree = read_xml(file_path=run_parameters_path)
        self.validate_instrument()

    def validate_instrument(self) -> None:
        """Raise an error if the parent class was instantiated."""
        raise NotImplementedError(
            "Parent class instantiated. Instantiate instead RunParametersHiSeq, "
            "RunParametersNovaSeq6000 or RunParametersNovaSeqX"
        )

    def _validate_instrument(self, node_name: str, node_value: str):
        """Fetches the node from an XML file and compares it with the expected value.
        Raises:
            RunParametersError if the node does not have the expected value."""
        try:
            application: ElementTree.Element | None = get_tree_node(
                tree=self.tree, node_name=node_name
            )
        except XMLError:
            raise RunParametersError(
                f"Could not find node {node_name} in the run parameters file. "
                "Check that the correct run parameters file is used"
            )
        if application.text != node_value:
            raise RunParametersError(f"The file parsed does not correspond to {node_value}")

    @property
    def index_length(self) -> int:
        """Return the length of the indexes if they are equal, raise an error otherwise."""
        index_one_length: int = self.get_index_1_cycles()
        index_two_length: int = self.get_index_2_cycles()
        if index_one_length != index_two_length and self.sequencer not in [
            Sequencers.HISEQX,
            Sequencers.HISEQGA,
        ]:
            raise RunParametersError("Index lengths are not the same!")
        return index_one_length

    def get_node_string_value(self, node_name: str) -> str:
        """Return the value of the node as a string if its validation passes."""
        return get_tree_node(tree=self.tree, node_name=node_name).text

    def get_node_integer_value(self, node_name: str) -> int:
        """Return the value of the node as an integer if its validation passes."""
        return int(self.get_node_string_value(node_name=node_name))

    @property
    @abstractmethod
    def control_software_version(self) -> str | None:
        """Return the control software version if existent."""
        pass

    @property
    @abstractmethod
    def reagent_kit_version(self) -> str | None:
        """Return the reagent kit version if existent."""
        pass

    @property
    @abstractmethod
    def sequencer(self) -> str | None:
        """Return the sequencer associated with the current run parameters."""
        pass

    @abstractmethod
    def get_index_1_cycles(self) -> int | None:
        """Return the number of cycles in the first index read."""
        pass

    @abstractmethod
    def get_index_2_cycles(self) -> int | None:
        """Return the number of cycles in the second index read."""
        pass

    @abstractmethod
    def get_read_1_cycles(self) -> int | None:
        """Return the number of cycles in the first read."""
        pass

    @abstractmethod
    def get_read_2_cycles(self) -> int | None:
        """Return the number of cycles in the second read."""
        pass

    def __str__(self):
        return f"RunParameters(path={self.path}, sequencer={self.sequencer})"

    def __repr__(self):
        return (
            f"RunParameters(path={self.path},"
            f"reagent_kit_version={self.reagent_kit_version},"
            "control_software_version={self.control_software_version},"
            f"index_length={self.index_length},"
            f"sequencer={self.sequencer})"
        )


class RunParametersHiSeq(RunParameters):
    """Specific class for parsing run parameters of HiSeq2500 sequencing."""

    def validate_instrument(self) -> None:
        """Validate if a HiSeq file was used to instantiate the class."""
        self._validate_instrument(
            node_name=RunParametersXMLNodes.APPLICATION_NAME,
            node_value=RunParametersXMLNodes.HISEQ_APPLICATION,
        )

    @property
    def control_software_version(self) -> None:
        """Return None for run parameters associated with HiSeq sequencing."""
        return

    @property
    def reagent_kit_version(self) -> None:
        """Return None for run parameters associated with HiSeq sequencing."""
        return

    @property
    def sequencer(self) -> str:
        """Return the sequencer associated with the current run parameters."""
        node_name: str = RunParametersXMLNodes.SEQUENCER_ID
        sequencer: str = self.get_node_string_value(node_name=node_name)
        return SEQUENCER_TYPES.get(sequencer)

    def get_index_1_cycles(self) -> int:
        """Return the number of cycles in the first index read."""
        node_name: str = RunParametersXMLNodes.INDEX_1_HISEQ
        return self.get_node_integer_value(node_name=node_name)

    def get_index_2_cycles(self) -> int:
        """Return the number of cycles in the second index read."""
        node_name: str = RunParametersXMLNodes.INDEX_2_HISEQ
        return self.get_node_integer_value(node_name=node_name)

    def get_read_1_cycles(self) -> int:
        """Return the number of cycles in the first read."""
        node_name: str = RunParametersXMLNodes.READ_1_HISEQ
        return self.get_node_integer_value(node_name=node_name)

    def get_read_2_cycles(self) -> int:
        """Return the number of cycles in the second read."""
        node_name: str = RunParametersXMLNodes.READ_2_HISEQ
        return self.get_node_integer_value(node_name=node_name)


class RunParametersNovaSeq6000(RunParameters):
    """Specific class for parsing run parameters of NovaSeq6000 sequencing."""

    def validate_instrument(self) -> None:
        """Validate if a NovaSeq6000 file was used to instantiate the class."""
        self._validate_instrument(
            node_name=RunParametersXMLNodes.APPLICATION,
            node_value=RunParametersXMLNodes.NOVASEQ_6000_APPLICATION,
        )

    @property
    def control_software_version(self) -> str:
        """Return the control software version."""
        node_name: str = RunParametersXMLNodes.APPLICATION_VERSION
        return self.get_node_string_value(node_name=node_name)

    @property
    def reagent_kit_version(self) -> str:
        """Return the reagent kit version if existent, return 'unknown' otherwise."""
        node_name: str = RunParametersXMLNodes.REAGENT_KIT_VERSION
        xml_node: ElementTree.Element | None = self.tree.find(node_name)
        if xml_node is None:
            LOG.warning("Could not determine reagent kit version")
            LOG.info("Set reagent kit version to 'unknown'")
            return RunParametersXMLNodes.UNKNOWN_REAGENT_KIT_VERSION
        return xml_node.text

    @property
    def sequencer(self) -> str:
        """Return the sequencer associated with the current run parameters."""
        return Sequencers.NOVASEQ

    def get_index_1_cycles(self) -> int:
        """Return the number of cycles in the first index read."""
        node_name: str = RunParametersXMLNodes.INDEX_1_NOVASEQ_6000
        return self.get_node_integer_value(node_name=node_name)

    def get_index_2_cycles(self) -> int:
        """Return the number of cycles in the second index read."""
        node_name: str = RunParametersXMLNodes.INDEX_2_NOVASEQ_6000
        return self.get_node_integer_value(node_name=node_name)

    def get_read_1_cycles(self) -> int:
        """Return the number of cycles in the first read."""
        node_name: str = RunParametersXMLNodes.READ_1_NOVASEQ_6000
        return self.get_node_integer_value(node_name=node_name)

    def get_read_2_cycles(self) -> int:
        """Return the number of cycles in the second read."""
        node_name: str = RunParametersXMLNodes.READ_2_NOVASEQ_6000
        return self.get_node_integer_value(node_name=node_name)


class RunParametersNovaSeqX(RunParameters):
    """Specific class for parsing run parameters of NovaSeqX sequencing."""

    def validate_instrument(self) -> None:
        """Validate if a NovaSeqX file was used to instantiate the class."""
        self._validate_instrument(
            node_name=RunParametersXMLNodes.INSTRUMENT_TYPE,
            node_value=RunParametersXMLNodes.NOVASEQ_X_INSTRUMENT,
        )

    @property
    def control_software_version(self) -> None:
        """Return None for run parameters associated with NovaSeqX sequencing."""
        return

    @property
    def reagent_kit_version(self) -> None:
        """Return None for run parameters associated with NovaSeqX sequencing."""
        return

    @property
    def sequencer(self) -> str:
        """Return the sequencer associated with the current run parameters."""
        return Sequencers.NOVASEQX

    @property
    def read_parser(self) -> dict[str, int]:
        """Return read and index cycle values parsed as a dictionary."""
        cycle_mapping: dict[str, int] = {}
        planned_reads_tree: ElementTree.Element = get_tree_node(
            tree=self.tree, node_name=RunParametersXMLNodes.PLANNED_READS_NOVASEQ_X
        )
        planned_reads: list[ElementTree.Element] = planned_reads_tree.findall(
            RunParametersXMLNodes.INNER_READ
        )
        for read_elem in planned_reads:
            read_name: str = read_elem.get(RunParametersXMLNodes.READ_NAME)
            cycles: int = int(read_elem.get(RunParametersXMLNodes.CYCLES))
            cycle_mapping[read_name] = cycles
        return cycle_mapping

    def get_index_1_cycles(self) -> int:
        """Return the number of cycles in the first index read."""
        return self.read_parser.get(RunParametersXMLNodes.INDEX_1_NOVASEQ_X)

    def get_index_2_cycles(self) -> int:
        """Return the number of cycles in the second index read."""
        return self.read_parser.get(RunParametersXMLNodes.INDEX_2_NOVASEQ_X)

    def get_read_1_cycles(self) -> int:
        """Return the number of cycles in the first read."""
        return self.read_parser.get(RunParametersXMLNodes.READ_1_NOVASEQ_X)

    def get_read_2_cycles(self) -> int:
        """Return the number of cycles in the second read."""
        return self.read_parser.get(RunParametersXMLNodes.READ_2_NOVASEQ_X)



"""Module for Balsamic Analysis API"""

import logging

from cg.constants import Pipeline
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


class BalsamicUmiAnalysisAPI(BalsamicAnalysisAPI):
    """Handles communication between BALSAMIC processes
    and the rest of CG infrastructure"""

    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.BALSAMIC_UMI,
    ):
        super().__init__(config=config, pipeline=pipeline)



import logging
import shutil
from pathlib import Path

from cg.constants import Pipeline
from cg.constants.constants import FileFormat
from cg.io.controller import WriteFile
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fastq import MutantFastqHandler
from cg.models.cg_config import CGConfig
from cg.models.workflow.mutant import MutantSampleConfig
from cg.store.models import Application, Case, Sample
from cg.utils import Process

LOG = logging.getLogger(__name__)


class MutantAnalysisAPI(AnalysisAPI):
    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.SARS_COV_2,
    ):
        super().__init__(config=config, pipeline=pipeline)
        self.root_dir = config.mutant.root

    @property
    def conda_binary(self) -> str:
        return self.config.mutant.conda_binary

    @property
    def process(self) -> Process:
        if not self._process:
            self._process = Process(
                binary=self.config.mutant.binary_path,
                conda_binary=f"{self.conda_binary}" if self.conda_binary else None,
                environment=self.config.mutant.conda_env,
            )
        return self._process

    @property
    def fastq_handler(self):
        return MutantFastqHandler

    @property
    def use_read_count_threshold(self) -> bool:
        return False

    def get_case_path(self, case_id: str) -> Path:
        return Path(self.root_dir, case_id)

    def get_case_output_path(self, case_id: str) -> Path:
        return Path(self.get_case_path(case_id=case_id), "results")

    def get_case_fastq_dir(self, case_id: str) -> Path:
        return Path(self.get_case_path(case_id=case_id), "fastq")

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        return Path(self.get_case_output_path(case_id=case_id), "trailblazer_config.yaml")

    def _is_nanopore(self, application: Application) -> bool:
        return application.tag[3:6] == "ONT"

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample) -> Path:
        """Return the path to the FASTQ destination directory."""
        application: str = sample.application_version.application
        if self._is_nanopore(application=application):
            return Path(self.get_case_path(case_id=case.internal_id), FileFormat.FASTQ, sample.name)
        return Path(self.get_case_path(case_id=case.internal_id), FileFormat.FASTQ)

    def get_case_config_path(self, case_id: str) -> Path:
        return Path(self.get_case_path(case_id=case_id), "case_config.json")

    def get_deliverables_file_path(self, case_id: str) -> Path:
        return Path(self.get_case_output_path(case_id=case_id), f"{case_id}_deliverables.yaml")

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        case_obj = self.status_db.get_case_by_internal_id(internal_id=case_id)
        samples: list[Sample] = [link.sample for link in case_obj.links]
        for sample_obj in samples:
            application = sample_obj.application_version.application
            if self._is_nanopore(application):
                self.link_nanopore_fastq_for_sample(
                    case_obj=case_obj, sample_obj=sample_obj, concatenate=True
                )
                continue
            if not sample_obj.sequencing_qc:
                LOG.info("Sample %s read count below threshold, skipping!", sample_obj.internal_id)
                continue
            else:
                self.link_fastq_files_for_sample(
                    case_obj=case_obj, sample_obj=sample_obj, concatenate=True
                )

    def get_sample_parameters(self, sample_obj: Sample) -> MutantSampleConfig:
        return MutantSampleConfig(
            CG_ID_sample=sample_obj.internal_id,
            case_ID=sample_obj.links[0].case.internal_id,
            Customer_ID_sample=sample_obj.name,
            customer_id=sample_obj.customer.internal_id,
            sequencing_qc_pass=sample_obj.sequencing_qc,
            CG_ID_project=sample_obj.links[0].case.name,
            Customer_ID_project=sample_obj.original_ticket,
            application_tag=sample_obj.application_version.application.tag,
            method_libprep=str(self.lims_api.get_prep_method(lims_id=sample_obj.internal_id)),
            method_sequencing=str(
                self.lims_api.get_sequencing_method(lims_id=sample_obj.internal_id)
            ),
            date_arrival=str(sample_obj.received_at),
            date_libprep=str(sample_obj.prepared_at),
            date_sequencing=str(sample_obj.last_sequenced_at),
            selection_criteria=self.lims_api.get_sample_attribute(
                lims_id=sample_obj.internal_id, key="selection_criteria"
            ),
            priority=self.lims_api.get_sample_attribute(
                lims_id=sample_obj.internal_id, key="priority"
            ),
            region_code=self.lims_api.get_sample_attribute(
                lims_id=sample_obj.internal_id, key="region_code"
            ),
            lab_code=self.lims_api.get_sample_attribute(
                lims_id=sample_obj.internal_id, key="lab_code"
            ),
            primer=self.lims_api.get_sample_attribute(lims_id=sample_obj.internal_id, key="primer"),
        )

    def create_case_config(self, case_id: str, dry_run: bool) -> None:
        case_obj = self.status_db.get_case_by_internal_id(internal_id=case_id)
        samples: list[Sample] = [link.sample for link in case_obj.links]
        case_config_list = [
            self.get_sample_parameters(sample_obj).model_dump() for sample_obj in samples
        ]
        config_path = self.get_case_config_path(case_id=case_id)
        if dry_run:
            LOG.info("Dry-run, would have created config at path %s, with content:", config_path)
            LOG.info(case_config_list)
        config_path.parent.mkdir(parents=True, exist_ok=True)
        WriteFile.write_file_from_content(
            content=case_config_list, file_format=FileFormat.JSON, file_path=config_path
        )
        LOG.info("Saved config to %s", config_path)

    def get_additional_naming_metadata(self, sample_obj: Sample) -> str | None:
        sample_name = sample_obj.name
        region_code = self.lims_api.get_sample_attribute(
            lims_id=sample_obj.internal_id, key="region_code"
        ).split(" ")[0]
        lab_code = self.lims_api.get_sample_attribute(
            lims_id=sample_obj.internal_id, key="lab_code"
        ).split(" ")[0]

        return f"{region_code}_{lab_code}_{sample_name}"

    def run_analysis(self, case_id: str, dry_run: bool, config_artic: str = None) -> None:
        if self.get_case_output_path(case_id=case_id).exists():
            LOG.info("Found old output files, directory will be cleaned!")
            shutil.rmtree(self.get_case_output_path(case_id=case_id), ignore_errors=True)

        if config_artic:
            self.process.run_command(
                [
                    "analyse",
                    "sarscov2",
                    "--config_case",
                    self.get_case_config_path(case_id=case_id).as_posix(),
                    "--config_artic",
                    config_artic,
                    "--outdir",
                    self.get_case_output_path(case_id=case_id).as_posix(),
                    self.get_case_fastq_dir(case_id=case_id).as_posix(),
                ],
                dry_run=dry_run,
            )
        else:
            self.process.run_command(
                [
                    "analyse",
                    "sarscov2",
                    "--config_case",
                    self.get_case_config_path(case_id=case_id).as_posix(),
                    "--outdir",
                    self.get_case_output_path(case_id=case_id).as_posix(),
                    self.get_case_fastq_dir(case_id=case_id).as_posix(),
                ],
                dry_run=dry_run,
            )

    def get_cases_to_store(self) -> list[Case]:
        """Return cases where analysis has a deliverables file,
        and is ready to be stored in Housekeeper."""
        return [
            case
            for case in self.status_db.get_running_cases_in_pipeline(pipeline=self.pipeline)
            if Path(self.get_deliverables_file_path(case_id=case.internal_id)).exists()
        ]

    def get_metadata_for_nanopore_sample(self, sample_obj: Sample) -> list[dict]:
        return [
            self.fastq_handler.parse_nanopore_file_data(file_obj.full_path)
            for file_obj in self.housekeeper_api.files(
                bundle=sample_obj.internal_id, tags=["fastq"]
            )
        ]

    def link_nanopore_fastq_for_sample(
        self, case_obj: Case, sample_obj: Sample, concatenate: bool = False
    ) -> None:
        """
        Link FASTQ files for a nanopore sample to working directory.
        If pipeline input requires concatenated fastq, files can also be concatenated
        """
        read_paths = []
        files: list[dict] = self.get_metadata_for_nanopore_sample(sample_obj=sample_obj)
        sorted_files = sorted(files, key=lambda k: k["path"])
        fastq_dir = self.get_sample_fastq_destination_dir(case=case_obj, sample=sample_obj)
        fastq_dir.mkdir(parents=True, exist_ok=True)

        for counter, fastq_data in enumerate(sorted_files):
            fastq_path = Path(fastq_data["path"])
            fastq_name = self.fastq_handler.create_nanopore_fastq_name(
                flowcell=fastq_data["flowcell"],
                sample=sample_obj.internal_id,
                filenr=str(counter),
                meta=self.get_additional_naming_metadata(sample_obj),
            )
            destination_path: Path = fastq_dir / fastq_name
            read_paths.append(destination_path)

            if not destination_path.exists():
                LOG.info(f"Linking: {fastq_path} -> {destination_path}")
                destination_path.symlink_to(fastq_path)
            else:
                LOG.warning(f"Destination path already exists: {destination_path}")

        if not concatenate:
            return

        concatenated_fastq_name = self.fastq_handler.create_nanopore_fastq_name(
            flowcell=fastq_data["flowcell"],
            sample=sample_obj.internal_id,
            filenr=str(1),
            meta=self.get_additional_naming_metadata(sample_obj),
        )
        concatenated_path = (
            f"{fastq_dir}/{self.fastq_handler.get_concatenated_name(concatenated_fastq_name)}"
        )
        LOG.info("Concatenation in progress for sample %s.", sample_obj.internal_id)
        self.fastq_handler.concatenate(read_paths, concatenated_path)
        self.fastq_handler.remove_files(read_paths)



import logging

from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.raredisease.base import raredisease
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_raredisease_no_args(cli_runner: CliRunner, raredisease_context: CGConfig):
    """Test to see that running RAREDISEASE without options prints help and doesn't result in an error."""
    # GIVEN no arguments or options besides the command call

    # WHEN running command
    result = cli_runner.invoke(raredisease, [], obj=raredisease_context)

    # THEN command runs successfully
    print(result.output)

    assert result.exit_code == EXIT_SUCCESS

    # THEN help should be printed
    assert "help" in result.output



"""Module for archiving and retrieving folders via DDN Dataflow."""
import logging
from datetime import datetime
from enum import StrEnum
from pathlib import Path
from urllib.parse import urljoin

from housekeeper.store.models import File
from pydantic import BaseModel, Field
from requests.models import Response

from cg.constants.constants import APIMethods
from cg.exc import DdnDataflowAuthenticationError
from cg.io.controller import APIRequest
from cg.meta.archive.models import (
    ArchiveHandler,
    FileAndSample,
    FileTransferData,
    SampleAndDestination,
)
from cg.models.cg_config import DataFlowConfig
from cg.store.models import Sample

LOG = logging.getLogger(__name__)

OSTYPE: str = "Unix/MacOS"
ROOT_TO_TRIM: str = "/home"

DESTINATION_ATTRIBUTE: str = "destination"
SOURCE_ATTRIBUTE: str = "source"


class DataflowEndpoints(StrEnum):
    """Enum containing all DDN dataflow endpoints used."""

    ARCHIVE_FILES = "files/archive"
    GET_AUTH_TOKEN = "auth/token"
    REFRESH_AUTH_TOKEN = "auth/token/refresh"
    RETRIEVE_FILES = "files/retrieve"
    GET_JOB_STATUS = "activity/jobs/"


class JobStatus(StrEnum):
    """Enum for the different job statuses which can be returned via Miria."""

    CANCELED = "Canceled"
    COMPLETED = "Completed"
    DENIED = "Denied"
    CREATION_IN_PROGRESS = "Creation in progress"
    IN_QUEUE = "In Queue"
    INVALID_LICENSE = "Invalid license"
    ON_VALIDATION = "On validation"
    REFUSED = "Refused"
    RUNNING = "Running"
    SUSPENDED = "Suspended"
    TERMINATED_ON_ERROR = "Terminated on error"
    TERMINATED_ON_WARNING = "Terminated on warning"


class MiriaObject(FileTransferData):
    """Model for representing a singular object transfer."""

    _metadata = None
    destination: str
    source: str

    @classmethod
    def create_from_file_and_sample(
        cls, file: File, sample: Sample, is_archiving: bool = True
    ) -> "MiriaObject":
        """Instantiates the class from a File and Sample object."""
        if is_archiving:
            return cls(destination=sample.internal_id, source=file.full_path)
        return cls(destination=file.full_path, source=sample.internal_id)

    @classmethod
    def create_from_sample_and_destination(
        cls, sample_and_destination: SampleAndDestination
    ) -> "MiriaObject":
        """Instantiates the class from a SampleAndDestination object,
        i.e. when we want to fetch a folder containing all spring files for said sample."""
        return cls(
            destination=sample_and_destination.destination,
            source=sample_and_destination.sample.internal_id,
        )

    def trim_path(self, attribute_to_trim: str):
        """Trims the given attribute (source or destination) from its root directory."""
        setattr(
            self,
            attribute_to_trim,
            f"/{Path(getattr(self, attribute_to_trim)).relative_to(ROOT_TO_TRIM)}",
        )

    def add_repositories(self, source_prefix: str, destination_prefix: str):
        """Prepends the given repositories to the source and destination paths."""
        self.source: str = source_prefix + self.source
        self.destination: str = destination_prefix + self.destination


class TransferPayload(BaseModel):
    """Model for representing a Dataflow transfer task."""

    files_to_transfer: list[MiriaObject]
    osType: str = OSTYPE
    createFolder: bool = False
    settings: list[dict] = []

    def trim_paths(self, attribute_to_trim: str):
        """Trims the source path from its root directory for all objects in the transfer."""
        for miria_file in self.files_to_transfer:
            miria_file.trim_path(attribute_to_trim=attribute_to_trim)

    def add_repositories(self, source_prefix: str, destination_prefix: str):
        """Prepends the given repositories to the source and destination paths all objects in the
        transfer."""
        for miria_file in self.files_to_transfer:
            miria_file.add_repositories(
                source_prefix=source_prefix, destination_prefix=destination_prefix
            )

    def model_dump(self, **kwargs) -> dict:
        """Creates a correctly structured dict to be used as the request payload."""
        payload: dict = super().model_dump(exclude={"files_to_transfer"})
        payload["pathInfo"] = [miria_file.model_dump() for miria_file in self.files_to_transfer]
        payload["metadataList"] = []
        return payload

    def post_request(self, url: str, headers: dict) -> "TransferJob":
        """Sends a request to the given url with, the given headers, and its own content as payload.

        Arguments:
            url: URL to which the POST goes to.
            headers: Headers which are set in the request
        Raises:
            HTTPError if the response status is not successful.
            ValidationError if the response does not conform to the expected response structure.
        Returns:
            The job ID of the launched transfer task.
        """

        response: Response = APIRequest.api_request_from_content(
            api_method=APIMethods.POST,
            url=url,
            headers=headers,
            json=self.model_dump(),
        )
        response.raise_for_status()
        return TransferJob.model_validate(response.json())


class AuthPayload(BaseModel):
    """Model representing the payload for an Authentication request."""

    dbName: str
    name: str
    password: str
    superUser: bool = False


class RefreshPayload(BaseModel):
    """Model representing the payload for Auth-token refresh request."""

    refresh: str


class AuthToken(BaseModel):
    """Model representing the response fields from an access request to the Dataflow API."""

    access: str
    expire: int
    refresh: str | None = None


class TransferJob(BaseModel):
    """Model representing th response fields of an archive or retrieve reqeust to the Dataflow
    API."""

    job_id: int = Field(alias="jobId")


class GetJobStatusResponse(BaseModel):
    """Model representing the response fields from a get_job_status post."""

    job_id: int = Field(alias="id")
    status: str


class GetJobStatusPayload(BaseModel):
    """Model representing the payload for a get_job_status request."""

    id: int

    def get_job_status(self, url: str, headers: dict) -> GetJobStatusResponse:
        """Sends a get request to the given URL with the given headers.
        Returns the parsed status response of the task specified in the URL.
        Raises:
             HTTPError if the response code is not successful."""
        response: Response = APIRequest.api_request_from_content(
            api_method=APIMethods.GET, url=url, headers=headers, json={}
        )
        response.raise_for_status()
        return GetJobStatusResponse.model_validate(response.json())


class DDNDataFlowClient(ArchiveHandler):
    """Class for archiving and retrieving folders via DDN Dataflow."""

    def __init__(self, config: DataFlowConfig):
        self.database_name: str = config.database_name
        self.user: str = config.user
        self.password: str = config.password
        self.url: str = config.url
        self.archive_repository: str = config.archive_repository
        self.local_storage: str = config.local_storage
        self.auth_token: str
        self.refresh_token: str
        self.token_expiration: datetime
        self.headers: dict[str, str] = {
            "Content-Type": "application/json",
            "accept": "application/json",
        }
        self._set_auth_tokens()

    def _set_auth_tokens(self) -> None:
        """Retrieves and sets auth and refresh token from the REST-API."""
        response: Response = APIRequest.api_request_from_content(
            api_method=APIMethods.POST,
            url=urljoin(base=self.url, url=DataflowEndpoints.GET_AUTH_TOKEN),
            headers=self.headers,
            json=AuthPayload(
                dbName=self.database_name,
                name=self.user,
                password=self.password,
            ).model_dump(),
        )
        if not response.ok:
            raise DdnDataflowAuthenticationError(message=response.text)
        response_content: AuthToken = AuthToken.model_validate(response.json())
        self.refresh_token: str = response_content.refresh
        self.auth_token: str = response_content.access
        self.token_expiration: datetime = datetime.fromtimestamp(response_content.expire)

    def _refresh_auth_token(self) -> None:
        """Updates the auth token by providing the refresh token to the REST-API."""
        response: Response = APIRequest.api_request_from_content(
            api_method=APIMethods.POST,
            url=urljoin(base=self.url, url=DataflowEndpoints.REFRESH_AUTH_TOKEN),
            headers=self.headers,
            json=RefreshPayload(refresh=self.refresh_token).model_dump(),
        )
        response_content: AuthToken = AuthToken.model_validate(response.json())
        self.auth_token: str = response_content.access
        self.token_expiration: datetime = datetime.fromtimestamp(response_content.expire)

    @property
    def auth_header(self) -> dict[str, str]:
        """Returns an authorization header based on the current auth token, or updates it if
        needed."""
        if datetime.now() > self.token_expiration:
            self._refresh_auth_token()
        return {"Authorization": f"Bearer {self.auth_token}"}

    def archive_files(self, files_and_samples: list[FileAndSample]) -> int:
        """Archives all files provided, to their corresponding destination, as given by sources
        and destination in TransferData. Returns the job ID of the archiving task."""
        miria_file_data: list[MiriaObject] = self.convert_into_transfer_data(
            files_and_samples, is_archiving=True
        )
        archival_request: TransferPayload = self.create_transfer_request(
            miria_file_data=miria_file_data, is_archiving_request=True
        )
        return archival_request.post_request(
            headers=dict(self.headers, **self.auth_header),
            url=urljoin(base=self.url, url=DataflowEndpoints.ARCHIVE_FILES),
        ).job_id

    def retrieve_samples(self, samples_and_destinations: list[SampleAndDestination]) -> int:
        """Retrieves all archived files for the provided samples and stores them in the specified location in
        Housekeeper."""
        miria_file_data: list[MiriaObject] = []
        for sample_and_housekeeper_destination in samples_and_destinations:
            miria_object: MiriaObject = MiriaObject.create_from_sample_and_destination(
                sample_and_housekeeper_destination
            )
            miria_file_data.append(miria_object)
        retrieval_request: TransferPayload = self.create_transfer_request(
            miria_file_data=miria_file_data, is_archiving_request=False
        )
        return retrieval_request.post_request(
            headers=dict(self.headers, **self.auth_header),
            url=urljoin(base=self.url, url=DataflowEndpoints.RETRIEVE_FILES),
        ).job_id

    def create_transfer_request(
        self, miria_file_data: list[MiriaObject], is_archiving_request: bool
    ) -> TransferPayload:
        """Performs the necessary curation of paths for the request to be valid, depending on if
        it is an archiving or a retrieve request.
        """
        source_prefix: str
        destination_prefix: str
        attribute: str

        source_prefix, destination_prefix, attribute = (
            (self.local_storage, self.archive_repository, SOURCE_ATTRIBUTE)
            if is_archiving_request
            else (self.archive_repository, self.local_storage, DESTINATION_ATTRIBUTE)
        )

        transfer_request: TransferPayload = TransferPayload(files_to_transfer=miria_file_data)
        transfer_request.trim_paths(attribute_to_trim=attribute)
        transfer_request.add_repositories(
            source_prefix=source_prefix, destination_prefix=destination_prefix
        )

        return transfer_request

    def convert_into_transfer_data(
        self, files_and_samples: list[FileAndSample], is_archiving: bool = True
    ) -> list[MiriaObject]:
        """Converts the provided files and samples to the format used for the request."""
        return [
            MiriaObject.create_from_file_and_sample(
                file=file_and_sample.file, sample=file_and_sample.sample, is_archiving=is_archiving
            )
            for file_and_sample in files_and_samples
        ]

    def is_job_done(self, job_id: int) -> bool:
        get_job_status_payload = GetJobStatusPayload(id=job_id)
        get_job_status_response: GetJobStatusResponse = get_job_status_payload.get_job_status(
            url=urljoin(self.url, DataflowEndpoints.GET_JOB_STATUS + str(job_id)),
            headers=dict(self.headers, **self.auth_header),
        )
        if get_job_status_response.status == JobStatus.COMPLETED:
            return True
        LOG.info(
            f"Job with id {job_id} has not been completed. "
            f"Current job description is {get_job_status_response.status}"
        )
        return False



import logging
from pathlib import Path

import click

from cg.constants.nipt import Q30_THRESHOLD
from cg.exc import AnalysisUploadError
from cg.meta.upload.nipt.nipt import NiptUploadAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.group()
def ftp():
    """Upload NIPT result files to ftp-server"""
    pass


@ftp.command("case")
@click.argument("case_id", required=True)
@click.option("--dry-run", is_flag=True)
@click.option("--force", is_flag=True, help="Force upload of case to databases, despite qc")
@click.pass_obj
def nipt_upload_case(context: CGConfig, case_id: str, dry_run: bool, force: bool):
    """Upload the results file of a NIPT case"""
    nipt_upload_api: NiptUploadAPI = NiptUploadAPI(context)
    nipt_upload_api.set_dry_run(dry_run=dry_run)

    if force or nipt_upload_api.flowcell_passed_qc_value(
        case_id=case_id, q30_threshold=Q30_THRESHOLD
    ):
        LOG.info("*** NIPT FTP UPLOAD START ***")

        hk_results_file: str = nipt_upload_api.get_housekeeper_results_file(case_id=case_id)
        results_file: Path = nipt_upload_api.get_results_file_path(hk_results_file)

        LOG.info(f"Results file found: {results_file}")
        LOG.info("Starting ftp upload!")

        nipt_upload_api.upload_to_ftp_server(results_file)

        LOG.info("Upload ftp finished!")
    else:
        LOG.error(f"Uploading case failed: {case_id}")
        LOG.error(
            f"Flowcell did not pass one of the following QC parameters:\n"
            f"target_reads={nipt_upload_api.target_reads(case_id=case_id)}, Q30_threshold={Q30_THRESHOLD}"
        )
        raise AnalysisUploadError("Upload failed")


@ftp.command("all")
@click.option("--dry-run", is_flag=True)
@click.pass_context
def nipt_upload_all(context: click.Context, dry_run: bool):
    """Upload all available NIPT results files"""

    LOG.info("*** UPLOAD ALL AVAILABLE NIPT RESULTS ***")

    nipt_upload_api = NiptUploadAPI(context.obj)
    analyses = nipt_upload_api.get_all_upload_analyses()
    if not analyses:
        LOG.info("No analyses found for upload")
        return

    for analysis in analyses:
        case_id = analysis.case.internal_id
        context.invoke(nipt_upload_case, case_id=case_id, dry_run=dry_run)



"""Module to handle dictionary helper functions."""


def get_list_from_dictionary(dictionary: dict) -> list:
    """Return a list of the passed dict non-empty key values."""
    list_from_dict: list[str] = []
    for key, value in dictionary.items():
        if value:
            list_from_dict.extend([key, value])
    return list_from_dict


def get_full_path_dictionary(dictionary: dict) -> dict:
    """Return a full path of the values of the passed dictionary."""
    full_path_dict: dict = {}
    for key, value in dictionary.items():
        full_path_dict[key] = value.full_path if value else None

    return full_path_dict



import pytest

from cg.exc import OrderError
from cg.meta.orders.sars_cov_2_submitter import SarsCov2Submitter
from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import ControlEnum
from cg.models.orders.samples import SarsCov2Sample
from cg.store import Store
from tests.store_helpers import StoreHelpers


def test_validate_normal_order(sarscov2_order_to_submit: dict, base_store: Store):
    # GIVEN sarscov2 order with three samples, none in the database
    order = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)

    # WHEN validating the order
    SarsCov2Submitter(status=base_store, lims=None).validate_order(order=order)
    # THEN it should be regarded as valid


def test_validate_submitted_order(
    sarscov2_order_to_submit: dict, base_store: Store, helpers: StoreHelpers
):
    # GIVEN sarscov2 order with three samples, all in the database
    order: OrderIn = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)

    sample: SarsCov2Sample
    for sample in order.samples:
        helpers.add_sample(store=base_store, name=sample.name, customer_id=order.customer)

    # WHEN validating the order
    # THEN it should be regarded as invalid
    with pytest.raises(OrderError):
        SarsCov2Submitter(status=base_store, lims=None).validate_order(order=order)


def test_validate_submitted_control_order(
    sarscov2_order_to_submit: dict, base_store: Store, helpers: StoreHelpers
):
    # GIVEN sarscov2 order with three control samples, all in the database
    order: OrderIn = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)

    sample: SarsCov2Sample
    for sample in order.samples:
        helpers.add_sample(store=base_store, name=sample.name, customer_id=order.customer)
        sample.control = ControlEnum.positive

    # WHEN validating the order
    # THEN it should be regarded as valid
    SarsCov2Submitter(status=base_store, lims=None).validate_order(order=order)



import datetime as dt
from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.housekeeper.models import InputBundle
from cg.constants import Pipeline
from cg.meta.workflow.fluffy import FluffyAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Sample
from tests.store_helpers import StoreHelpers


@pytest.fixture(scope="function")
def fluffy_case_id_existing():
    return "norwegiangiraffe"


@pytest.fixture(scope="function")
def fluffy_case_id_non_existing():
    return "nakedmolerat"


@pytest.fixture(scope="function")
def fluffy_sample_lims_id():
    return "ACC1234A1"


@pytest.fixture(scope="function")
def fluffy_success_output_summary(tmpdir_factory):
    output_dir = tmpdir_factory.mktemp("output")
    file_path = Path(output_dir, "summary.csv")
    file_path.touch(exist_ok=True)
    return file_path


@pytest.fixture(scope="function")
def fluffy_success_output_aberrations(tmpdir_factory):
    output_dir = tmpdir_factory.mktemp("output")
    file_path = Path(output_dir, "WCXpredict_aberrations.filt.bed")
    file_path.touch(exist_ok=True)
    return file_path


@pytest.fixture
def bcl_convert_samplesheet_path() -> Path:
    return Path("tests", "fixtures", "data", "bcl_convert_sample_sheet.csv")


@pytest.fixture
def sample() -> Sample:
    return Sample(
        name="sample_name",
        order="sample_project",
        control="positive",
        last_sequenced_at=dt.datetime.now(),
    )


@pytest.fixture(scope="function")
def samplesheet_path():
    return Path("tests/fixtures/data/SampleSheet.csv").absolute()


@pytest.fixture(scope="function")
def fluffy_fastq_file_path(config_root_dir):
    path = Path(config_root_dir)
    path.mkdir(parents=True, exist_ok=True)
    fastq_path = Path(path, "fastq.fastq.gz")
    fastq_path.touch(exist_ok=True)
    return fastq_path


@pytest.fixture(scope="function")
def deliverables_yaml_path():
    return Path("tests/fixtures/apps/fluffy/deliverables.yaml")


@pytest.fixture
def fluffy_hermes_deliverables_response_data(
    create_multiqc_html_file,
    fluffy_case_id_existing,
    fluffy_sample_lims_id,
    fluffy_success_output_summary,
    fluffy_success_output_aberrations,
    timestamp_yesterday,
):
    return InputBundle(
        **{
            "files": [
                {
                    "path": fluffy_success_output_summary.as_posix(),
                    "tags": ["metrics", fluffy_case_id_existing, "nipt"],
                },
                {
                    "path": create_multiqc_html_file.as_posix(),
                    "tags": ["multiqc-html", fluffy_case_id_existing, "nipt"],
                },
                {
                    "path": fluffy_success_output_aberrations.as_posix(),
                    "tags": ["wisecondor", "cnv", fluffy_sample_lims_id, "nipt"],
                },
            ],
            "created": timestamp_yesterday,
            "name": fluffy_case_id_existing,
        }
    )


@pytest.fixture(scope="function")
def fluffy_fastq_hk_bundle_data(fluffy_fastq_file_path, fluffy_sample_lims_id) -> dict:
    return {
        "name": fluffy_sample_lims_id,
        "created": dt.datetime.now(),
        "version": "1.0",
        "files": [
            {
                "path": fluffy_fastq_file_path.as_posix(),
                "tags": ["fastq", "flowcell"],
                "archive": False,
            }
        ],
    }


@pytest.fixture(scope="function")
def fluffy_samplesheet_bundle_data(samplesheet_path) -> dict:
    return {
        "name": "flowcell",
        "created": dt.datetime.now(),
        "version": "1.0",
        "files": [
            {
                "path": str(samplesheet_path),
                "tags": ["flowcell", "samplesheet"],
                "archive": False,
            }
        ],
    }


@pytest.fixture(scope="function")
def fluffy_context(
    cg_context: CGConfig,
    helpers: StoreHelpers,
    real_housekeeper_api: HousekeeperAPI,
    fluffy_samplesheet_bundle_data,
    fluffy_fastq_hk_bundle_data,
    fluffy_case_id_existing,
    fluffy_sample_lims_id,
) -> CGConfig:
    cg_context.housekeeper_api_ = real_housekeeper_api
    fluffy_analysis_api = FluffyAnalysisAPI(config=cg_context)
    helpers.ensure_hk_version(
        fluffy_analysis_api.housekeeper_api, bundle_data=fluffy_samplesheet_bundle_data
    )
    helpers.ensure_hk_version(fluffy_analysis_api.housekeeper_api, fluffy_fastq_hk_bundle_data)
    example_fluffy_case = helpers.add_case(
        fluffy_analysis_api.status_db,
        internal_id=fluffy_case_id_existing,
        name=fluffy_case_id_existing,
        data_analysis=Pipeline.FLUFFY,
    )
    example_fluffy_sample = helpers.add_sample(
        fluffy_analysis_api.status_db,
        internal_id=fluffy_sample_lims_id,
        is_tumour=False,
        application_type="tgs",
        reads=100,
        last_sequenced_at=dt.datetime.now(),
    )
    helpers.add_flow_cell(
        fluffy_analysis_api.status_db, flow_cell_name="flowcell", samples=[example_fluffy_sample]
    )
    helpers.add_relationship(
        fluffy_analysis_api.status_db, case=example_fluffy_case, sample=example_fluffy_sample
    )
    cg_context.meta_apis["analysis_api"] = fluffy_analysis_api
    return cg_context



from sqlalchemy.orm import Query

from cg.store import Store
from cg.store.filters.status_case_sample_filters import (
    get_cases_with_sample_by_internal_id,
    get_samples_in_case_by_internal_id,
)


def test_get_samples_in_case_by_internal_id_valid_id(
    store_with_analyses_for_cases: Store,
    case_id: str,
):
    """Test that filtering by a valid case id returns case-samples with the desired case id."""
    # GIVEN a store with case-samples and a case internal id
    case_sample_query: Query = store_with_analyses_for_cases._get_join_case_sample_query()

    # WHEN filtering by the chosen case internal id
    filtered_query: Query = get_samples_in_case_by_internal_id(
        case_samples=case_sample_query, case_internal_id=case_id
    )

    # THEN the filtered query has at least one element but fewer elements than the original query
    assert 0 < filtered_query.count() < case_sample_query.count()
    # THEN the case_samples in the filtered query have the correct case internal id
    for case_sample in filtered_query.all():
        assert case_sample.case.internal_id == case_id


def test_get_samples_in_case_by_internal_id_nonexistent_id(
    store_with_analyses_for_cases: Store, case_id_does_not_exist: str
):
    """Test that an empty query is returned when filtering using a non-existent case id."""
    # GIVEN a store with case-samples and an unexistent case internal id
    case_sample_query: Query = store_with_analyses_for_cases._get_join_case_sample_query()
    assert case_sample_query.count() > 0

    # WHEN filtering using a non-existent id
    filtered_query: Query = get_samples_in_case_by_internal_id(
        case_samples=case_sample_query, case_internal_id=case_id_does_not_exist
    )

    # THEN the filtered query is empty
    assert filtered_query.count() == 0


def test_get_cases_with_sample_by_internal_id_valid_id(
    store_with_analyses_for_cases: Store,
    sample_id: str,
):
    """Test that filtering by a valid sample internal id returns case-samples with the desired sample internal id."""
    # GIVEN a store with case-samples and a sample internal id
    case_sample_query: Query = store_with_analyses_for_cases._get_join_case_sample_query()

    # WHEN filtering using a sample internal id
    filtered_query: Query = get_cases_with_sample_by_internal_id(
        case_samples=case_sample_query, sample_internal_id=sample_id
    )

    # THEN the filtered query has at least one element but fewer elements than the original query
    assert 0 < filtered_query.count() < case_sample_query.count()
    # THEN the case_samples in the filtered query have the correct sample internal id
    for case_sample in filtered_query.all():
        assert case_sample.sample.internal_id == sample_id


def test_get_cases_with_sample_by_internal_id_invalid_id(
    store_with_analyses_for_cases: Store, invalid_sample_id: str
):
    """Test that an empty query is returned when filtering using an invalid sample internal id."""
    # GIVEN a store with case-samples and an invalid sample internal id
    case_sample_query: Query = store_with_analyses_for_cases._get_join_case_sample_query()
    assert case_sample_query.count() > 0

    # WHEN filtering using an invalid sample internal id
    filtered_query: Query = get_cases_with_sample_by_internal_id(
        case_samples=case_sample_query, sample_internal_id=invalid_sample_id
    )

    # THEN the filtered query is empty
    assert filtered_query.count() == 0



from enum import StrEnum

from cg.constants.constants import Pipeline


class OrderType(StrEnum):
    BALSAMIC: str = str(Pipeline.BALSAMIC)
    BALSAMIC_QC: str = str(Pipeline.BALSAMIC_QC)
    BALSAMIC_UMI: str = str(Pipeline.BALSAMIC_UMI)
    FASTQ: str = str(Pipeline.FASTQ)
    FLUFFY: str = str(Pipeline.FLUFFY)
    METAGENOME: str = "metagenome"
    MICROSALT: str = str(Pipeline.MICROSALT)
    MIP_DNA: str = str(Pipeline.MIP_DNA)
    MIP_RNA: str = str(Pipeline.MIP_RNA)
    RML: str = "rml"
    RNAFUSION: str = str(Pipeline.RNAFUSION)
    SARS_COV_2: str = str(Pipeline.SARS_COV_2)


class ExcelSampleAliases(StrEnum):
    AGE_AT_SAMPLING = "UDF/age_at_sampling"
    APPLICATION = "UDF/Sequencing Analysis"
    CAPTURE_KIT = "UDF/Capture Library version"
    COHORTS = "UDF/cohorts"
    COLLECTION_DATE = "UDF/Collection Date"
    COMMENT = "UDF/Comment"
    CONCENTRATION = "UDF/Concentration (nM)"
    CONCENTRATION_SAMPLE = "UDF/Sample Conc."
    CONTAINER = "Container/Type"
    CONTAINER_NAME = "Container/Name"
    CONTROL = "UDF/Control"
    CUSTOM_INDEX = "UDF/Custom index"
    CUSTOMER = "UDF/customer"
    DATA_ANALYSIS = "UDF/Data Analysis"
    DATA_DELIVERY = "UDF/Data Delivery"
    ELUTION_BUFFER = "UDF/Sample Buffer"
    EXTRACTION_METHOD = "UDF/Extraction method"
    FAMILY_NAME = "UDF/familyID"
    FATHER = "UDF/fatherID"
    FORMALIN_FIXATION_TIME = "UDF/Formalin Fixation Time"
    INDEX = "UDF/Index type"
    INDEX_NUMBER = "UDF/Index number"
    LAB_CODE = "UDF/Lab Code"
    MOTHER = "UDF/motherID"
    NAME = "Sample/Name"
    ORGANISM = "UDF/Strain"
    ORGANISM_OTHER = "UDF/Other species"
    ORIGINAL_LAB = "UDF/Original Lab"
    ORIGINAL_LAB_ADDRESS = "UDF/Original Lab Address"
    PANELS = "UDF/Gene List"
    POOL = "UDF/pool name"
    POST_FORMALIN_FIXATION_TIME = "UDF/Post Formalin Fixation Time"
    PRE_PROCESSING_METHOD = "UDF/Pre Processing Method"
    PRIMER = "UDF/Primer"
    PRIORITY = "UDF/priority"
    QUANTITY = "UDF/Quantity"
    REAGENT_LABEL = "Sample/Reagent Label"
    REFERENCE_GENOME = "UDF/Reference Genome Microbial"
    REGION = "UDF/Region"
    REGION_CODE = "UDF/Region Code"
    REQUIRE_QC_OK = "UDF/Process only if QC OK"
    RML_PLATE_NAME = "UDF/RML plate name"
    SELECTION_CRITERIA = "UDF/Selection Criteria"
    SEX = "UDF/Gender"
    SOURCE = "UDF/Source"
    STATUS = "UDF/Status"
    SUBJECT_ID = "UDF/subjectID"
    SYNOPSIS = "UDF/synopsis"
    TISSUE_BLOCK_SIZE = "UDF/Tissue Block Size"
    TUMOUR = "UDF/tumor"
    TUMOUR_PURITY = "UDF/tumour purity"
    VOLUME = "UDF/Volume (uL)"
    WELL_POSITION = "Sample/Well Location"
    WELL_POSITION_RML = "UDF/RML well position"



"""CLI support to create config and/or start TAXPROFILER."""

import logging

import click
from pydantic.v1 import ValidationError

from cg.cli.workflow.commands import ARGUMENT_CASE_ID, resolve_compression
from cg.cli.workflow.nf_analysis import (
    OPTION_COMPUTE_ENV,
    OPTION_CONFIG,
    OPTION_LOG,
    OPTION_PARAMS_FILE,
    OPTION_PROFILE,
    OPTION_REVISION,
    OPTION_TOWER_RUN_ID,
    OPTION_USE_NEXTFLOW,
    OPTION_WORKDIR,
)
from cg.cli.workflow.taxprofiler.options import (
    OPTION_FROM_START,
    OPTION_INSTRUMENT_PLATFORM,
)
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.constants.constants import DRY_RUN, CaseActions, MetaApis
from cg.constants.nf_analysis import NfTowerStatus
from cg.constants.sequencing import SequencingPlatform
from cg.exc import CgError, DecompressionNeededError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.taxprofiler import TaxprofilerAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.rnafusion.rnafusion import CommandArgs

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.pass_context
def taxprofiler(context: click.Context) -> None:
    """nf-core/taxprofiler analysis workflow."""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis[MetaApis.ANALYSIS_API] = TaxprofilerAnalysisAPI(
        config=context.obj,
    )


taxprofiler.add_command(resolve_compression)


@taxprofiler.command("config-case")
@ARGUMENT_CASE_ID
@OPTION_INSTRUMENT_PLATFORM
@DRY_RUN
@click.pass_obj
def config_case(
    context: CGConfig, case_id: str, instrument_platform: SequencingPlatform, dry_run: bool
) -> None:
    """Create sample sheet and parameter file for Taxprofiler analysis for a given case."""
    analysis_api: TaxprofilerAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    try:
        analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
        analysis_api.config_case(
            case_id=case_id, instrument_platform=instrument_platform, dry_run=dry_run
        )

    except (CgError, ValidationError) as error:
        LOG.error(f"Could not create config files for {case_id}: {error}")
        raise click.Abort() from error


@taxprofiler.command("run")
@ARGUMENT_CASE_ID
@OPTION_LOG
@OPTION_WORKDIR
@OPTION_FROM_START
@OPTION_PROFILE
@OPTION_CONFIG
@OPTION_PARAMS_FILE
@OPTION_REVISION
@OPTION_COMPUTE_ENV
@OPTION_USE_NEXTFLOW
@OPTION_TOWER_RUN_ID
@DRY_RUN
@click.pass_obj
def run(
    context: CGConfig,
    case_id: str,
    log: str,
    work_dir: str,
    from_start: bool,
    profile: str,
    config: str,
    params_file: str,
    revision: str,
    compute_env: str,
    use_nextflow: bool,
    nf_tower_id: str | None,
    dry_run: bool,
) -> None:
    """Run taxprofiler analysis for a case."""
    analysis_api: TaxprofilerAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)

    command_args: CommandArgs = CommandArgs(
        **{
            "log": analysis_api.get_log_path(
                case_id=case_id,
                pipeline=analysis_api.pipeline,
                log=log,
            ),
            "work_dir": analysis_api.get_workdir_path(case_id=case_id, work_dir=work_dir),
            "resume": not from_start,
            "profile": analysis_api.get_profile(profile=profile),
            "config": analysis_api.get_nextflow_config_path(nextflow_config=config),
            "params_file": analysis_api.get_params_file_path(
                case_id=case_id, params_file=params_file
            ),
            "name": case_id,
            "compute_env": compute_env or analysis_api.compute_env,
            "revision": revision or analysis_api.revision,
            "wait": NfTowerStatus.SUBMITTED,
            "id": nf_tower_id,
        }
    )
    try:
        analysis_api.verify_sample_sheet_exists(case_id=case_id, dry_run=dry_run)
        analysis_api.check_analysis_ongoing(case_id)
        LOG.info(f"Running Taxprofiler analysis for {case_id}")
        analysis_api.run_analysis(
            case_id=case_id, command_args=command_args, use_nextflow=use_nextflow, dry_run=dry_run
        )
        analysis_api.set_statusdb_action(
            case_id=case_id, action=CaseActions.RUNNING, dry_run=dry_run
        )
    except Exception as error:
        LOG.error(f"Could not run analysis: {error}")
        raise click.Abort() from error
    if not dry_run and not use_nextflow:
        analysis_api.add_pending_trailblazer_analysis(case_id=case_id)


@taxprofiler.command("start")
@ARGUMENT_CASE_ID
@OPTION_LOG
@OPTION_WORKDIR
@OPTION_PROFILE
@OPTION_CONFIG
@OPTION_PARAMS_FILE
@OPTION_REVISION
@OPTION_COMPUTE_ENV
@OPTION_USE_NEXTFLOW
@OPTION_TOWER_RUN_ID
@DRY_RUN
@click.pass_context
def start(
    context: CGConfig,
    case_id: str,
    log: str,
    work_dir: str,
    profile: str,
    config: str,
    params_file: str,
    revision: str,
    compute_env: str,
    use_nextflow: bool,
    nf_tower_id: str | None,
    dry_run: bool,
) -> None:
    """Start full workflow for case id."""
    LOG.info(f"Starting analysis for {case_id}")

    try:
        context.invoke(resolve_compression, case_id=case_id, dry_run=dry_run)
    except DecompressionNeededError as error:
        LOG.error(error)
        raise click.Abort() from error
    context.invoke(config_case, case_id=case_id, dry_run=dry_run)
    context.invoke(
        run,
        case_id=case_id,
        log=log,
        work_dir=work_dir,
        from_start=True,
        profile=profile,
        nf_tower_id=nf_tower_id,
        config=config,
        params_file=params_file,
        revision=revision,
        compute_env=compute_env,
        use_nextflow=use_nextflow,
        dry_run=dry_run,
    )


@taxprofiler.command("start-available")
@DRY_RUN
@click.pass_context
def start_available(context: click.Context, dry_run: bool = False) -> None:
    """Start full workflow for all cases available for analysis."""

    analysis_api: AnalysisAPI = context.obj.meta_apis[MetaApis.ANALYSIS_API]

    exit_code: int = EXIT_SUCCESS
    for case in analysis_api.get_cases_to_analyze():
        try:
            context.invoke(start, case_id=case.internal_id, dry_run=dry_run)
        except Exception as error:
            LOG.error(error)
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort
